"""
Traffic provides :manpage:`kqueue(2)` based event driven I/O for Python.
However, in addition to listening for the I/O events, Traffic performs the
corresponding I/O operation prior to propagating the notification.
In essence, Traffic transforms can-read and can-write events into did-read and
did-write.

.. warning:: Traffic a new project. The APIs may change without warning.

Traffic is designed to maximize efficiency by performing I/O operations in bulk.
By grouping I/O operations with event collections, Traffic is capable of
performing a significant amount of work while the GIL is released, giving the
process the CPU time necessary to consume the data transfers performed by the
prior cycle.

Traffic aims to be a foundation for higher-level I/O management APIs, so
direct use of Traffic often requires a fair amount of boilerplate code. However,
some tools are provided to make direct use easier.

Transferring Data with Traffic
------------------------------

Making a connection is easy::

	from traffic import libkqueue as libtraffic, libloop
	s1, s2 = libtraffic.Octets.connect_stream_socketpair()

Traffic manages each channel as a separate object, so ``s1`` and ``s2`` is
actually a :py:class:`tuple` of two :py:class:`traffic.libkqueue.Octets`. The receive
part is the first item and the send part, the second. However, in order for
these objects to perform actual transfers, they must be attached to a
:py:class:`traffic.libkqueue.Traffic` instance::

	traffic = libtraffic.Traffic()
	s1read, s1write = s1
	s2read, s2write = s2

Attachment is performed with the :py:meth:`traffic.libkqueue.Traffic.acquire` method.
Once acquired, the Transit is owned by the Traffic object, and will not be
released until termination::

	traffic.acquire(s1read)
	traffic.acquire(s1write)
	traffic.acquire(s2read)
	traffic.acquire(s2write)

Now that the traffic instance has acquired the Transits, transfers are possible
given resource availability: the Transits now need to acquire resources to
transfer; some data to send or some space to receive into. Similar to Traffic,
this is done with the acquire method, :py:meth:`traffic.libkqueue.Octets.acquire`::

	mutable_buffer = libtraffic.Octets.rallocate(128) # memoryview(bytearray()) object
	s1read.acquire(mutable_buffer)
	s1write.acquire(b'Hello, that side of the world!')

Likewise the other end needs to acquire resources as well::

	s2read.acquire(libtraffic.Octets.rallocate(128))
	s2write.acquire(b'Nobody is home! Go Away!')

All of the Transits have the necessary resources for performing a transfer.
However, the transfers cannot be performed until the Traffic loop can respond
to all the events--resource acquisition is an event from the perspective of a
Traffic instance::

	import queue
	q = queue.Queue(6)

Once there is a destination queue, the loop can be kicked off. In this example, a
:py:class:`threading.Thread` instance is used to run the loop::

	import threading
	deliver = q.put
	thread = threading.Thread(args = (deliver, traffic), target = libloop.loop)
	thread.start()

The loop is now performing transfers and delivering sequences of
:py:class:`traffic.libloop.Activity` instances using :py:meth:`q.put`.
Reading the sequences with :py:meth:`q.get`, the transfers can be processed
while the :py:func:`traffic.libloop.loop` performs the next set of transfers for
subsequent processing::

	io = q.get()
	while io is not None:
		for x in io:
			if x.demand is not None:
				# It's a write Transit that has sent all its data.
				x.transit.terminate()
			else:
				print(x)
		if traffic.volume == 1:
			# all transits have terminated
			traffic.terminate()
		io = q.get()
	print("Complete!")

The output of the above should look something like::

	Activity(transit=<traffic.libkqueue.Octets object at 0xDEADBEEF1>, link=None, termination=None, transferred=bytearray(b'Hello, that side of the world!'), demand=None)
	Activity(transit=<traffic.libkqueue.Octets object at 0xDEADBEEF2>, link=None, termination=None, transferred=bytearray(b'Nobody is home! Go Away!'), demand=None)
	Activity(transit=<traffic.libkqueue.Octets object at 0xDEADBEEF3>, link=None, termination=(None, 0), transferred=None, demand=None)
	Activity(transit=<traffic.libkqueue.Octets object at 0xDEADBEEF4>, link=None, termination=(None, 0), transferred=None, demand=None)
	Activity(transit=<traffic.libkqueue.Octets object at 0xDEADBEEF5>, link=None, termination=('kevent', 0), transferred=None, demand=None)
	Activity(transit=<traffic.libkqueue.Octets object at 0xDEADBEEF6>, link=None, termination=('kevent', 0), transferred=None, demand=None)

Naturally, the code in this example doesn't really benefit from having the
loop performed in a separate thread. However, the purpose of the example is to
illuminate the anticipated form of use. In cases of numerous connections, the above
model is extremely efficient as transfers are *actually* being performed
concurrently with processing. The concurrency itself does not contrast much with
traditional multi-threaded Python processes. Most Python interfaces to
underlying C functions release the GIL and allow true concurrency with regards
to the processing performed by the C function. The contrast is in the form of GIL
contention: Traffic loops can perform *multiple* I/O operations without the GIL
being held.

The use of :py:mod:`traffic.libloop` is not required, but it does make a few
things easier as the Traffic APIs avoid higher-level concepts in order to
encourage adaption to framework-local APIs.

For more information, the Usage chapter of the documentation covers bare-use
of :py:mod:`traffic.lib` as well as :py:mod:`traffic.libloop`.
"""
__pkg_bottom__ = True
