/*
 * libkqueue.py.c - kernel/kqueue based traffic
 *
 * Transit ("abstract" base)
 *  |- Point (abstract base for kpoint (file descriptor) based Transits)
 *   |- Octets (transfers "octets", bytes of data (read/write))
 *   |- Sockets (transfers "sockets", file descriptors (listen/accept))
 *   |- Traffic (transfers Transits that have a non-pending event)
 *
 *  Conventions and Terminology:
 *  	'T' (capital t) refers to the Traffic instance.
 *  	't' (lowercase t) refers to an arbitrary Transit instance.
 *  	'o' (lowercase o) refers to an arbitrary Octets instance.
 *  	's' (lowercase s) refers to an arbitrary Sockets instance.
 *
 *  	'equals' refers to Event QUALificationS
 *  	 These are state flags that must be present in order for a
 *  	 particular event to come to pass.
 *  	'polarity' refers to the direction of the transit; sends or receives.
 *  	'qualify' refers to setting a flag on 'equals'
 *  	'nqualify' refers to removing a flag on 'equals'
 *  	'qualified' refers to testing for the presence of a flag on 'equals'
 *  	'qual prefixed with "x"' refers to external 'equals' (kernel)
 *  	'qual prefixed with "i"' refers to internal 'equals' (process)
 *
 * The traffic/documentation/mechanics.txt file contains more information.
 */
#define PY_SSIZE_T_CLEAN
#include <Python.h>
#include <structmember.h>
#include <pythread.h>
#define PyObject_TypeName(ob) ((Py_TYPE(ob))->tp_name)
#define NONE() Py_None; Py_INCREF(Py_None)

#ifndef HAVE_STDINT_H
#include <stdint.h>
#endif

#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <netinet/in.h>

/* kqueue */
#include <sys/event.h>

#ifdef _LINUX_SOCKET_H
#define SETADDRLEN(TARGET, VALUE)
#else
#define SETADDRLEN(TARGET, VALUE) TARGET = VALUE
#endif

#define DEFAULT_BACKLOG 64
#define DEFAULT_BACKLOG_STR "64"

/* Number of kevent structs to allocate when working with kevent(). */
#ifndef N_KEVENTS
#define N_KEVENTS 32
#endif

#ifndef PACKAGE_NAME
#define PACKAGE_NAME "traffic"
#endif

#ifndef PACKAGE
#define PACKAGE PACKAGE_NAME "."
#endif

#ifndef MODULE_NAME
#define MODULE_NAME "libkqueue"
#endif

#ifndef MODULE
#define MODULE PACKAGE MODULE_NAME
#endif

#define QPATH(S) MODULE "." S

#if PY_MAJOR_VERSION > 2
#define MODULE_ENTRY_POINT(NAME) PyInit_##NAME
#else
#define MODULE_ENTRY_POINT(NAME) init##NAME
#endif

/* Documentation Reference: Class */
#define DRC(Y) ":py:class:`" QPATH(#Y) "`"
#define DRM(Y) ":py:meth:`" QPATH(#Y) "`"

#if 0
#define PRINTF_DEBUGGING
#endif

#define MIN(x1,x2) ((x1) < (x2) ? (x1) : (x2))

typedef PyObject * PyObj;
typedef int kpoint_t; /* file descriptor */
typedef int kerror_t; /* kerror error identifier (errno) */
typedef uint8_t kcall_t; /* the contrived kcall type */
typedef struct kevent *kevent_t; /* kernel event description */
typedef int (*kp_buffer_op_t)(kpoint_t fd, void *buf, int amount);

/*
 * cpp list of syscalls used by traffic.libkqueue
 */
#define KCALLS() \
	KC(kqueue) \
	KC(kevent) \
	KC(read) \
	KC(write) \
	KC(sendmsg) \
	KC(recvmsg) \
	KC(setsockopt) \
	KC(getsockopt) \
	KC(fcntl) \
	KC(socketpair) \
	KC(pipe) \
	KC(open) \
	KC(lseek) \
	KC(socket) \
	KC(bind) \
	KC(connect) \
	KC(listen) \
	KC(shutdown) \
	KC(accept) \
	KC(close) \
	KC(dup) \
	KC(dup2)

enum {
	kc_none = 0,
#define KC(x) kc_##x ,
	KCALLS()
#undef KC
	kc_terminal_
};

#define KC(x) kc_##x##_ob = NULL,
PyObj
KCALLS()
	kc_none_ob = NULL;
#undef KC
;

static PyObj
kcall_strob(kcall_t kc)
{
	PyObj rob;

	switch (kc)
	{
#define KC(x) case ((uint8_t) kc_##x): rob = kc_##x##_ob ; break;
		KCALLS()
#undef KC
		default:
			rob = Py_None;
		break;
	}

	Py_INCREF(rob);
	return(rob);
}

/*
 * Types of transits used here.
 */
typedef enum {
	transit_type = 0,	/* 't' */
	point_type,			/* 'p' */
	traffic_type,		/* 'T' */
	octets_type,		/* 'o' */
	sockets_type,		/* 'S' */
} transit_type_t;

static PyObj new_array = NULL; /* array.array("i", [-1]).__mul__ */
static PyObj PyExc_TransitError = NULL
	,PyExc_TerminatedError = NULL
	,PyExc_ResourceError = NULL
;
#define PyErr_SetTransitTerminatedError(t) \
	PyErr_SetString(PyExc_TerminatedError, Py_TYPE(t)->tp_name)
#define TRANSIT_EXCEPTIONS() \
	ID(TransitError) \
	ID(TerminatedError) \
	ID(ResourceError) \

static PyObj
returnself_noargs(PyObj self)
{
	Py_INCREF(self);
	return(self);
}

static int
inet6_from_pyint(void *out, PyObj ob)
{
	if (ob == Py_None)
		return(INADDR_ANY);
	return(_PyLong_AsByteArray((PyLongObject *) ob, out, 128 / 8, 0, 0));
}

static int
inet4_from_pyint(void *out, PyObj ob)
{
	if (ob == Py_None)
		return(INADDR_ANY);
	return(_PyLong_AsByteArray((PyLongObject *) ob, out, 32 / 8, 0, 0));
}

static int
init_kpoint(kpoint_t kp)
{
#ifdef F_SETNOSIGPIPE
	/*
	 * No sigpipe.
	 */
	fcntl(kp, F_SETNOSIGPIPE, 1);
	errno = 0;
#endif

	/*
	 * This must succeed.
	 */
	if (fcntl(kp, F_SETFL, O_NONBLOCK))
		return(-1);

	return(0);
}

static int
socket_rcvbuf(kpoint_t kp)
{
	int size = 0;
	socklen_t ssize = sizeof(size);
	getsockopt(kp, SOL_SOCKET, SO_RCVBUF, &size, &ssize);
	return(size);
}

static int
socket_sndbuf(kpoint_t kp)
{
	int size = -1;
	socklen_t ssize = sizeof(size);
	getsockopt(kp, SOL_SOCKET, SO_SNDBUF, &size, &ssize);
	return(size);
}

/*
 * init_socket - setsockopt's and run init_kpoint
 */
static int
init_socket(kpoint_t kp)
{
	static int true_indicator = 1;

	setsockopt(kp, SOL_SOCKET, SO_KEEPALIVE, &true_indicator, sizeof(true_indicator));

	/*
	 * Inline out of band data. Octets only support "single channel".
	 * Most protocols don't use OOB and to avoid situations where OOB data fills
	 * up our buffer, inline it and make the user handle it in stream.
	 */
	setsockopt(kp, SOL_SOCKET, SO_OOBINLINE, &true_indicator, sizeof(true_indicator));
#ifndef F_SETNOSIGPIPE
#ifdef SO_NOSIGPIPE
	setsockopt(kp, SOL_SOCKET, SO_NOSIGPIPE, &true_indicator, sizeof(true_indicator));
#endif
#endif

	return(init_kpoint(kp));
}

/*
 * Generalized events used by all Traffic Points.
 *
 * If changed, update traffic.libloop as well.
 * (Allows traffic.libloop to be tested without traffic.libkqueue dependency.)
 */
typedef enum {
/* No event */
/* Events emitted by Transits */
	tev_terminate = 0, /* transit was terminated */
	tev_transfer,  /* transfer occurred */
	tev_exhaust,  /* end of resource */
	tev_terminal_,
	tev_traffic, /* internal event used to signal xconnect */
} transit_event_t;

#define TRANSIT_EVENTS() \
	ID(terminate) \
	ID(transfer) \
	ID(exhaust)

/*
 * event qualifications
 *
 * Pairs of qualifications must occur in order to produce an event.
 */
typedef enum {
	teq_polarity = 0, /* direction of the transit */
	teq_terminate, /* terminated noted/requested */
	teq_transfer, /* transfer potential noted */
	teq_terminal_,
} transit_equal_t; /* Event qualification */

#define TEQ_INTERNAL 0
#define TEQ_EXTERNAL 1

#define Transit_Equals(t) (t->junction)
#define Transit_AddEquals(t, equals) (Transit_Equals(t) |= equals)
#define Transit_IAddress(id) (1 << ((TEQ_INTERNAL * teq_terminal_) + id))
#define Transit_XAddress(id) (1 << ((TEQ_EXTERNAL * teq_terminal_) + id))
/* Used to detect event availability. */
#define Transit_Termination (Transit_IAddress(teq_terminate)|Transit_XAddress(teq_terminate))
#define Transit_Transferrence (Transit_IAddress(teq_transfer)|Transit_XAddress(teq_transfer))
#define Transit_DQualify(t, id) (Transit_GetDelta(t) |= Transit_IAddress(id))
#define Transit_DQualified(t, id) (Transit_GetDelta(t) & Transit_IAddress(id))

#define Transit_IQualified(t, id) (Transit_Equals(t) & Transit_IAddress(id))
#define Transit_IQualify(t, id) (Transit_Equals(t) |= Transit_IAddress(id))
#define Transit_INQualify(t, id) (Transit_Equals(t) ^= Transit_IAddress(id))
#define Transit_XQualified(t, id) (Transit_XAddress(id) & Transit_Equals(t))
#define Transit_XQualify(t, id) (Transit_Equals(t) |= Transit_XAddress(id))
#define Transit_XNQualify(t, id) (Transit_Equals(t) ^= Transit_XAddress(id))

#define Transit_Terminated(t) ((Transit_Equals(t) & Transit_Termination) == Transit_Termination)
#define Transit_Terminating(t) ( \
	(Transit_GetDelta(t) & Transit_IAddress(teq_terminate)) || \
	Transit_IQualified(t, teq_terminate) || \
	Transit_XQualified(t, teq_terminate) \
)

/*
 * If internal termination is requested do it or
 * If external termination occurred, but only when there is no xtransfer.
 */
#define Transit_ShouldTerminate(t) \
	(Transit_IQualified(t, teq_terminate) || (Transit_XQualified(t, teq_terminate)))
#define Transit_ShouldTransfer(t) ((Transit_Equals(t) & Transit_Transferrence) == Transit_Transferrence)
#define Transit_ShouldXConnect(t) \
	(Transit_HasEvent(t, tev_traffic) && !Transit_Terminated(t))

/* Initialized by the module entry point. */
static PyObj traffic_event_objects[tev_terminal_];
#define TRAFFIC_EVENT_OB(EID) (traffic_event_objects[(int) EID])

#define Transit_HEAD \
	PyObject_HEAD \
	Traffic traffic; /* transit controller */ \
	Transit prev, next; /* transit ring; all transit in traffic */ \
	Transit lltransfer; /* linked list pointer to next evented Transit */ \
	PyObj link; /* user field; normally a callback */ \
	\
	uint8_t delta; /* delta to apply to the junction; new internal equals (GIL) */ \
	/* */ \
	uint8_t junction; /* bit map of equals (event qualifications) */ \
	/* */ \
	uint8_t event; /* bit map of produced events */ \

typedef struct Traffic *Traffic;
typedef struct Transit *Transit;
struct Transit {
	Transit_HEAD
};

/*
 * TransitInterface methods for a given type.
 */
struct TransitInterface {
	/* identifies the kevent filter; NULL if not driven by kevent */
	void (*ti_kfilter)(Transit, kevent_t);

	/* transform kevent into Transit state */
	void (*ti_transform)(Transit, kevent_t);

	/* transferred units into the resource */
	void (*ti_transfer)(Transit);

	/* terminates the associated *representation* of the transit */
	void (*ti_terminate)(Transit);

	void (*ti_xrelease)(Transit);

	/* release internal resources */
	void (*ti_irelease)(Transit);
};
typedef struct TransitInterface *TransitInterface;
/*
 * TransitInterface is used internally by Traffic to perform transfers.
 * tif.* are the interfaces used by libtraffic,
 * whereas typ.* are the interfaces used by python.
 */
typedef struct TransitPyTypeObject {
	PyTypeObject typ;
	struct TransitInterface *tif;
} TransitPyTypeObject;

#define Transit_GetInterface(OB) (((TransitPyTypeObject *) Py_TYPE(OB))->tif)

#define Point_HEAD \
	Transit_HEAD \
	kpoint_t kpoint; \
	kcall_t kcall; /* the kernel entry point that caused the error */ \
	kerror_t kerror; /* error returned by system */ \
	Py_ssize_t ksize; /* "units" contained/available in kernel buffers. */

struct Point {
	Point_HEAD
};
typedef struct Point *Point;

/*
 * the kqueue() manager;
 * a set of transits (the transit transit)
 */
struct Traffic {
	Point_HEAD
	Py_ssize_t transit_count; /* total number of acquired transits */
	Py_ssize_t transfer_count; /* total number of transits in lltransfer */

	/*
	 * last transits in LL's; for fast appends.
	 */
	Transit lltransfer_ptr;

	/*
	 * Traffic's Resource; ksize is number of changes or collected.
	 */
	struct kevent kevs[N_KEVENTS];
};

typedef enum {
	octets_bad = 0, /* bad file descriptor */
	octets_pipe, /* anonymous pipe */
	octets_fifo, /* named pipe */
	octets_socket /* local or otherwise */,
	octets_file /* "regular" file */,
} octets_ktype_t;

static octets_ktype_t
point_ktype(kpoint_t p, blksize_t *blksize)
{
	struct stat st;
	if (fstat((int) p, &st) != 0)
		return octets_bad;

	*blksize = st.st_blksize;

	switch (st.st_mode & S_IFMT)
	{
		case S_IFSOCK:
			return octets_socket;
		break;
		case S_IFIFO:
			return octets_fifo;
		break;
		case S_IFREG:
			return octets_file;
		break;
	}

	/* assume pipe */
	return(octets_pipe);
}

/*
 * Transfers (read/write) octets to and from file descriptors
 */
typedef struct Octets *Octets;
struct Octets { /* represents the remote endpoint */
	Point_HEAD
	octets_ktype_t ktype;
	Py_ssize_t kcapacity; /* Size of kernel buffer */
	Octets opposite; /* not-NULL in the case of sockets */

	/* Octet buffer and state. If reading, buffer *must* be mutable. */
	Py_ssize_t position, last_position;
	PyObj resource; /* object with used buffer interface */
	Py_buffer view;
};

/*
 * Transfers socket file descriptors from a listening socket.
 */
struct Sockets {
	Point_HEAD
	Py_ssize_t position, last_position; /* the position in the buffer */
	PyObj resource; /* memoryview (!readonly) */
	Py_buffer view;
};
typedef struct Sockets *Sockets;

#define LLIter_HEAD \
	PyObject_HEAD \
	Traffic T; \
	Transit t; /* state */

struct lltransferiter {
	LLIter_HEAD
};

struct eventiter {
	LLIter_HEAD
	transit_event_t current_event;
};

#define PyAllocate(TYP) (((PyTypeObject *) TYP)->tp_alloc((PyTypeObject *) TYP, 0))

#define INIT_TRANSIT(t) do { \
	(t)->traffic = NULL; \
	(t)->junction = 0; (t)->delta = 0; (t)->event = 0; \
	(t)->link = NULL; (t)->lltransfer = NULL; \
} while (0)

#define Transit_GetTraffic(t) (t->traffic)
#define Transit_SetTraffic(t, T) (Transit_GetTraffic(t) = T)

#define Transit_HasEvent(T, TEV) (T->event & (1 << TEV))
#define Transit_NoteEvent(t, ev) (t->event |= (1 << ev))
#define Transit_RemoveEvent(T, TEV) (T->event ^= (1 << TEV))
#define Transit_ClearEvents(t) (t->event = 0)

#define Transit_GetJunction(t) (t->junction)
#define Transit_GetEvent(t) (t->event)
#define Transit_GetDelta(t) (t->delta)
#define Transit_ClearDelta(t) (Transit_GetDelta(t) = 0)
#define Transit_InCycle(t) (Transit_GetTraffic(t)->lltransfer != NULL)
#define Traffic_IncrementTransitCount(T) (++(T->transit_count))
#define Traffic_DecrementTransitCount(T) (--(T->transit_count))

#define Transit_Polarity(t) (Transit_IQualified(t, teq_polarity))
#define Transit_Sends(t) (!Transit_Polarity(t))
#define Transit_Receives(t) (Transit_Polarity(t))

/* Append to the end of the linked list; requires GIL. */
#define Transit_EnqueueDelta(t) do { \
	if (Transit_GetDelta(t) != 0 && ((Traffic)(t)) != Transit_GetTraffic(t)) { \
		TRANSIT_RELOCATE_SEGMENT_BEFORE(Transit_GetTraffic(t), (t), (t)); \
	} \
} while(0)

#define Transit_EnqueueEvent(_t_) do { \
	/* Only append event if it's not already in lltransfer. */ \
	if (_t_->lltransfer == NULL) { \
		Traffic T = _t_->traffic; \
		_t_->lltransfer = (Transit) T; /* designates ll participation */ \
		T->lltransfer_ptr->lltransfer = _t_; \
		T->lltransfer_ptr = _t_; \
		++(T->transfer_count); \
	} \
} while(0)

/*
 * Traffic only holds one reference to transits no matter
 * how often they're referenced.
 */
#define TRANSIT_JOIN(PREV, NEXT) do { \
	PREV->next = (Transit) NEXT; \
	NEXT->prev = (Transit) PREV; \
} while (0)

/*
 * Extends ring from behind.
 */
#define TRANSIT_ATTACH_SEGMENT_BEFORE(TARGET, FIRST, LAST) do { \
	FIRST->prev = TARGET->prev; \
	LAST->next = (Transit) TARGET; \
	TARGET->prev->next = (Transit) FIRST; \
	TARGET->prev = (Transit) LAST; \
} while (0)

/*
 * Extends ring from front.
 */
#define TRANSIT_ATTACH_SEGMENT_AFTER(TARGET, FIRST, LAST) do { \
	FIRST->prev = (Transit) TARGET; \
	LAST->next = (Transit) TARGET->next; \
	TARGET->next->prev = (Transit) LAST; \
	TARGET->next = (Transit) FIRST; \
} while (0)

#define TRANSIT_RELOCATE_SEGMENT_AFTER(TARGET, HEAD, TAIL) do { \
	TRANSIT_DETACH_SEGMENT(HEAD, TAIL); \
	TRANSIT_ATTACH_SEGMENT_AFTER(TARGET, HEAD, TAIL); \
} while (0)

#define TRANSIT_RELOCATE_SEGMENT_BEFORE(TARGET, HEAD, TAIL) do { \
	TRANSIT_DETACH_SEGMENT(HEAD, TAIL); \
	TRANSIT_ATTACH_SEGMENT_BEFORE(TARGET, HEAD, TAIL); \
} while (0)

#define TRANSIT_DETACH_SEGMENT(HEAD, TAIL) do { \
	HEAD->prev->next = (Transit) TAIL->next; \
	TAIL->next->prev = (Transit) HEAD->prev; \
} while(0)

#define TRANSIT_DETACH(TRANSIT) do { \
	TRANSIT_DETACH_SEGMENT(TRANSIT, TRANSIT); \
	TRANSIT->prev = NULL; TRANSIT->next = NULL; \
} while(0)
#define TRANSIT_ATTACH_BEFORE(TARGET, TRANSIT) \
	TRANSIT_ATTACH_SEGMENT_BEFORE(TARGET, TRANSIT, TRANSIT)
#define TRANSIT_ATTACH_AFTER(TARGET, TRANSIT) \
	TRANSIT_ATTACH_SEGMENT_AFTER(TARGET, TRANSIT, TRANSIT)
#define TRANSIT_ATTACH(TRANSIT) \
	TRANSIT_ATTACH_BEFORE(Transit_GetTraffic(TRANSIT), TRANSIT)

#define Point_GetKPoint(P) (P->kpoint)
#define Point_SetKPoint(P, VAL) (Point_GetKPoint(P) = VAL)
#define Point_KPointIsValid(P) (Point_GetKPoint(P) >= 0)
#define Point_GetKError(P) (P->kerror)
#define Point_SetKError(P, VAL) (Point_GetKError(P) = VAL)
#define Point_GetKCall(P) (P->kcall)
#define Point_SetKCall(P, VAL) (Point_GetKCall(P) = VAL)
#define Point_SetErrno(P) (Point_SetKError(P, errno))
#define Point_NoteKError(P, VAL) (Point_GetKError(P) == 0 ? Point_SetKError(P, VAL) : Point_GetKError(P))

#define INIT_POINT(OB, KP, KC, KE) do { \
	Point _P_ = ((Point) OB); \
	INIT_TRANSIT(_P_); \
	Point_SetKPoint(_P_, KP); \
	Point_SetKCall(_P_, KC); \
	Point_SetKError(_P_, KE); \
	if (KE) {Transit_DQualify(_P_, teq_terminate); Transit_XQualify(_P_, teq_terminate);} \
} while(0)

#define Sockets_GetArray(S) ((int *)(((intptr_t) S->view.buf) + (S->position * sizeof(int))))
#define Sockets_GetRemainder(S) ((S->view.len / sizeof(int)) - S->position)

#define Octets_GetOpposite(o) (o->opposite)
#define Octets_SetOpposite(o, x) (Octets_GetOpposite(o) = x)
#define Octets_GetKType(o) (o->ktype)
#define Octets_SetKType(o, kt) (Octets_GetKType(o) = kt)
#define Octets_IsSocket(o) (Octets_GetKType(o) == octets_socket)
#define Octets_Mechanism(o) \
	((Transit_Sends(o) ? (kp_buffer_op_t) write : (kp_buffer_op_t) read))

#define INIT_OCTETS(OB, KPOINT, KC, KE) do { \
	Octets _o = (Octets) OB; \
	INIT_POINT(OB, KPOINT, KC, KE); \
	Octets_SetKType(_o, octets_pipe); \
	Octets_SetOpposite(_o, NULL); \
	_o->kcapacity = 0; _o->ksize = 0; \
	_o->position = -1; _o->last_position = -1; \
	_o->resource = NULL; \
} while(0)

#define INIT_SOCKETS(OB, KP, KC, KE) do { \
	Sockets _S = (Sockets) OB; \
	INIT_POINT(OB, KP, KC, KE); \
	Transit_IQualify(_S, teq_polarity); \
	_S->ksize = 0; \
	_S->position = _S->last_position = -1; \
	_S->resource = NULL; \
} while(0)

#define INIT_READ_OCTETS(o, KPOINT, KC, KE) do { \
	INIT_OCTETS(o, KPOINT, KC, KE); \
	Transit_IQualify(o, teq_polarity); \
	o->kcapacity = (Py_ssize_t) socket_rcvbuf(KPOINT); \
} while (0)

#define INIT_WRITE_OCTETS(o, KPOINT, KC, KE) do { \
	INIT_OCTETS(o, KPOINT, KC, KE); \
	Transit_XQualify(o, teq_polarity); /* X-Polarity Unused */ \
	o->kcapacity = (Py_ssize_t) socket_sndbuf(KPOINT); \
} while (0)

#define INIT_PAIR(IOB, OOB, IKP, OKP, KC, KE) do { \
	INIT_READ_OCTETS(IOB, IKP, KC, KE); \
	INIT_WRITE_OCTETS(OOB, OKP, KC, KE); \
	if (IKP == OKP) { /* same fd? it's a socket/shared fd. */ \
		Octets_SetKType(OOB, octets_socket); \
		Octets_SetKType(IOB, octets_socket); \
		Octets_SetOpposite(IOB, OOB); Py_INCREF(OOB); \
		Octets_SetOpposite(OOB, IOB); /* recv's dealloc will release OOB's ref */ \
	} \
} while (0)

#define INIT_PAIR_TUPLE(TUP, IKP, OKP, KC, KE) do { \
	Octets _INPUT, _OUTPUT; \
	_INPUT = (Octets) PyTuple_GET_ITEM(TUP, 0); \
	_OUTPUT = (Octets) PyTuple_GET_ITEM(TUP, 1); \
	INIT_PAIR(_INPUT, _OUTPUT, IKP, OKP, KC, KE); \
} while (0)

#ifdef PRINTF_DEBUGGING
static PyObject *
pkevent(kevent_t kev)
{
	const char *fname;
#define EV_FILTS \
	F(EVFILT_USER) \
	F(EVFILT_READ) \
	F(EVFILT_WRITE) \
	F(EVFILT_PROC) \
	F(EVFILT_VNODE) \
	F(EVFILT_AIO) \
	F(EVFILT_SIGNAL) \
	F(EVFILT_TIMER)
#define F(B) case B: fname = #B; break;
	switch (kev->filter)
	{
		EV_FILTS
		default:
			fname = "unknown filter";
		break;
	}

#define EV_FLAGS \
	XF(EV_ADD), \
	XF(EV_ENABLE), \
	XF(EV_DISABLE), \
	XF(EV_DELETE), \
	XF(EV_RECEIPT), \
	XF(EV_ONESHOT), \
	XF(EV_CLEAR), \
	XF(EV_EOF), \
	XF(EV_ERROR)
#define XF(FLG) (kev->flags & FLG) ? (#FLG "|") : ""

	printf(
		"%s (%d), fflags: %d,"
		" ident: %p, data: %p, udata: %p, flags:"
		" %s%s%s%s%s%s%s%s%s\n",
		fname, kev->filter, kev->fflags,
		(void *) kev->ident,
		(void *) kev->data, (void *) kev->udata,
		EV_FLAGS
	);
	return(NULL);
#undef XF
#undef EV_FLAGS
}

static void
ptransit(Transit t)
{
	printf("\t[refcnt:%zd] %s @ %p/%p<->:%p:<->%p, errno: %s, junction: %s%s%s%s%s%s, events: %s%s%s%s\n",
		Py_REFCNT(t),
		Py_TYPE(t)->tp_name, t->traffic, t->prev, t, t->next,
		Point_GetKError(t) != 0 ? strerror(Point_GetKError(t)) : "0",
		Transit_IQualified(t, teq_polarity) ? "IRECEIVES" : "ISENDS",
		Transit_IQualified(t, teq_terminate) ? "|ITerm" : "",
		Transit_IQualified(t, teq_transfer) ? "|ITransfer" : "",
		Transit_XQualified(t, teq_terminate) ? "|XTerm" : "",
		Transit_XQualified(t, teq_transfer) ? "|XTransfer" : "",
		Transit_XQualified(t, teq_polarity) ? "|XRECEIVES" : "|XSENDS",
		Transit_HasEvent(t, tev_exhaust) ? "|exhaust" : "",
		Transit_HasEvent(t, tev_terminate) ? "|terminate" : "",
		Transit_HasEvent(t, tev_transfer) ? "|transfer" : "",
		Transit_HasEvent(t, tev_traffic) ? "|traffic" : "",
		""
	);
}

static void
pkevents(const char *where, Traffic T)
{
	int i;
	printf("%s [\n", where);
	for (i = 0; i<T->ksize; ++i)
	{
		pkevent(&(T->kevs[i]));
		ptransit((Transit) (T->kevs[i]).udata);
	}
	printf("]\n");
}
#else
#define pkevents(...)
#define ptransit(...)
#endif

/*
 * Implementation
 */

static void
ti_transit_kfilter(Transit t, kevent_t kev)
{
	/*
	 * Fill in the struct kevent structure with the
	 * necessary information to add, delete, enable, or disable
	 * the corresponding filter.
	 */
}

static void
ti_transit_transform(Transit t, kevent_t kev)
{
	/*
	 * Transform the kevent into local transit state preparing
	 * for transfer or termination.
	 */
	return;
}

static void
ti_transit_transfer(Transit t)
{
	/*
	 * Perform a transfer based on the lastest state.
	 */
	return;
}

static void
ti_transit_terminate(Transit t)
{
	/*
	 * Destroy the representation associated with the transit.
	 * For example, shutdown() should be called on stream sockets.
	 */
	Transit_XQualify(t, teq_terminate);
	return;
}

static void
ti_transit_xrelease(Transit t)
{
	/*
	 * Free any kernel resources associated with the transit (file descriptors).
	 */
	return;
}

static void
ti_transit_irelease(Transit t)
{
	/*
	 * Free any Python resources associated with the transit.
	 */
	return;
}

struct TransitInterface TransitTIF = {
	ti_transit_kfilter,
	ti_transit_transform,
	ti_transit_transfer,
	ti_transit_terminate,
	ti_transit_xrelease,
	ti_transit_irelease,
};

static PyObj
lltransferiter_next(PyObj self)
{
	PyObj this;
	struct lltransferiter *i = (struct lltransferiter *) self;

	if (i->t == NULL || i->t == (Transit) i->T)
		return(NULL);

	if (!Transit_InCycle(i->t))
	{
		PyErr_SetString(PyExc_RuntimeError,
			"traffic transfer iterator used outside of cycle");
		return(NULL);
	}

	this = (PyObj) i->t; /* returns this reference */
	i->t = i->t->lltransfer;

	if (i->t == (Transit) i->T)
	{
		Py_DECREF(i->T);
		i->t = NULL;
		i->T = NULL;
	}
	else
		Py_INCREF(i->t);

	return(this);
}

static void
lliter_dealloc(PyObj self)
{
	struct lltransferiter *i = (struct lltransferiter *) self;
	Py_XDECREF(i->T);
	Py_XDECREF(i->t);
	i->T = NULL;
	i->t = NULL;
}

const char lliter_doc[] = "iterator for transits with events";
PyTypeObject lltransferiter_type = {
	PyVarObject_HEAD_INIT(NULL, 0)
	QPATH("lltransferiter"),	/* tp_name */
	sizeof(struct lltransferiter),/* tp_basicsize */
	0,									/* tp_itemsize */
	lliter_dealloc,				/* tp_dealloc */
	NULL,								/* tp_print */
	NULL,								/* tp_getattr */
	NULL,								/* tp_setattr */
	NULL,								/* tp_compare */
	NULL,								/* tp_repr */
	NULL,								/* tp_as_number */
	NULL,								/* tp_as_sequence */
	NULL,								/* tp_as_mapping */
	NULL,								/* tp_hash */
	NULL,								/* tp_call */
	NULL,								/* tp_str */
	NULL,								/* tp_getattro */
	NULL,								/* tp_setattro */
	NULL,								/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT,	  		/* tp_flags */
	lliter_doc,						/* tp_doc */
	NULL,								/* tp_traverse */
	NULL,								/* tp_clear */
	NULL,								/* tp_richcompare */
	0,									/* tp_weaklistoffset */
	returnself_noargs,			/* tp_iter */
	lltransferiter_next,			/* tp_iternext */
	NULL,
};

static PyObj
new_lltransfer_iter(Traffic T)
{
	struct eventiter *i;

	i = (struct eventiter *) PyAllocate(&lltransferiter_type);
	if (i == NULL)
		return(NULL);

	i->T = T;
	i->t = T->lltransfer;
	Py_XINCREF(i->t);
	Py_INCREF(i->T);

	return((PyObj) i);
}

static PyObj
evnext(PyObj self)
{
	PyObj ev = NULL;
	struct eventiter *i = (struct eventiter *) self;

	if (i->current_event == tev_terminal_)
		return(NULL); /* no more events */

	if (!Transit_InCycle(i->t))
	{
		PyErr_SetString(PyExc_RuntimeError, "event iterator used outside of cycle");
		return(NULL);
	}

	do
	{
		if (Transit_HasEvent(i->t, i->current_event))
		{
			ev = TRAFFIC_EVENT_OB(i->current_event);
			Py_INCREF(ev);
			++i->current_event;
			return(ev);
		}
		++i->current_event;
	} while (i->current_event < tev_terminal_);

	return(NULL);
}

const char eventiter_doc[] = "iterator for the events that occurred on a transit";
PyTypeObject eventiter_type = {
	PyVarObject_HEAD_INIT(NULL, 0)
	"traffic.libkqueue.eventiter",		/* tp_name */
	sizeof(struct eventiter),	/* tp_basicsize */
	0,									/* tp_itemsize */
	lliter_dealloc,				/* tp_dealloc */
	NULL,								/* tp_print */
	NULL,								/* tp_getattr */
	NULL,								/* tp_setattr */
	NULL,								/* tp_compare */
	NULL,								/* tp_repr */
	NULL,								/* tp_as_number */
	NULL,								/* tp_as_sequence */
	NULL,								/* tp_as_mapping */
	NULL,								/* tp_hash */
	NULL,								/* tp_call */
	NULL,								/* tp_str */
	NULL,								/* tp_getattro */
	NULL,								/* tp_setattro */
	NULL,								/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT,	  		/* tp_flags */
	eventiter_doc,					/* tp_doc */
	NULL,								/* tp_traverse */
	NULL,								/* tp_clear */
	NULL,								/* tp_richcompare */
	0,									/* tp_weaklistoffset */
	returnself_noargs,			/* tp_iter */
	evnext,							/* tp_iternext */
	NULL,
};

static PyObj
new_event_iter(Transit t)
{
	struct eventiter *i;

	i = (struct eventiter *) PyAllocate(&eventiter_type);
	if (i == NULL)
		return(NULL);

	i->T = t->traffic;
	i->t = t;
	if (i->t->event == 0)
		i->current_event = tev_terminal_;
	else
		i->current_event = 0;

	Py_INCREF(i->t);
	Py_INCREF(i->T);

	return((PyObj) i);
}

static char
transit_can_acquire(Transit t)
{
	if (Transit_Terminating(t))
	{
		PyErr_SetTransitTerminatedError(t);
		return(0);
	}

	/*
	 * This should be called after receiving an exhaust event, which
	 * removes this internal flag.
	 */
	if (Transit_IQualified(t, teq_transfer))
	{
		/*
		 * This needs to error out as the traffic flow may be using the
		 * transit's resource at this particular moment.
		 */
		PyErr_SetString(PyExc_ResourceError, "transit resource is busy");
		return(0);
	}

	return(1);
}

static void
transit_acquired(Transit t)
{
	if (Transit_GetTraffic(t) != NULL)
	{
		Transit_DQualify(t, teq_transfer);
		Transit_EnqueueDelta(t); /* REQUIRES GIL */
	}
	else
	{
		/* not in traffic, directly apply the event qualification */
		Transit_IQualify(t, teq_transfer);
	}
}

static PyObj
transit_terminate(PyObj self)
{
	Transit t = (Transit) self;

	if (Transit_GetTraffic(t) == NULL)
	{
		/*
		 * Has GIL, not in Traffic.
		 * Traffic instances cannot acquire Transits without the GIL.
		 *
		 * Running terminate directly is safe.
		 */
		if (!Transit_Terminated(t))
		{
			TransitInterface tif;
			tif = Transit_GetInterface(t);
			tif->ti_terminate(t);
			tif->ti_xrelease(t);
		}
	}
	else if (!Transit_IQualified(t, teq_terminate))
	{
		/*
		 * Acquired by a Traffic instance, that Traffic
		 * instance is responsible for running the ti_terminate().
		 *
		 * Has GIL, so place teq_terminate event qualification on the delta.
		 */
		Transit_DQualify(t, teq_terminate);
		Transit_EnqueueDelta(t); /* REQUIRES GIL */
	}

	Py_INCREF(self);
	return(self);
}

static PyObj
transit_disregard(PyObj self)
{
	Transit t = (Transit) self;

	if (!Transit_Terminating(t))
		Transit_GetInterface(t)->ti_xrelease(t);

	Py_INCREF(self);
	return(self);
}

static PyObj
transit_failure(PyObj self)
{
	Py_INCREF(Py_None);
	return(Py_None);
}

static PyObj
transit_events(PyObj self)
{
	return(new_event_iter((Transit) self));
}

static PyObj
transit_acquire(PyObj self)
{
	Py_INCREF(self);
	return(self);
}

static PyMethodDef transit_methods[] = {

	{"acquire",
		(PyCFunction) transit_acquire, METH_NOARGS,
		PyDoc_STR(
"acquire(resource)\n"
":param resource: The resource to use facitate transfers.\n"
":type resource: object\n"
"\n"
":returns: The Transit instance acquiring the resource. (Method Chaining)\n"
":rtype: " DRC(Transit) "\n"
"\n"
"Acquire a `resource` for facilitating transfers. The `resource` type depends on\n"
"the Transit subclass, but it is *normally* an object supporting the buffer interface.\n"
"The particular Transit type should document the kind of object it expects."
)},

	{"terminate",
		(PyCFunction) transit_terminate, METH_NOARGS,
		PyDoc_STR(
"terminate()\n\n"
":returns: The transit being terminated. (Method Chaining)\n"
":rtype: " DRC(Transit) "\n"
"\n"
"Terminate the Transit permanently causing events to subside. Eventually, \n"
"resources being held by the Tranist will be released."
)},

	{"disregard",
		(PyCFunction) transit_disregard, METH_NOARGS,
		PyDoc_STR(
"disregard()\n\n"
":returns: The transit being disregarded. (Method Chaining)\n"
":rtype: " DRC(Transit) "\n"
"\n"
"Deallocate references to resources held by the Transit.\n"
"The semantics are distinct from " DRM(terminate)
" as " DRM(disregard) " doesn't attempt to end the represented channel."
)},

	{"failure", (PyCFunction) transit_failure,
		METH_NOARGS, PyDoc_STR(
"failure()\n\n"
":returns: An object describing the failure if any.\n"
":rtype: :py:class:`object`\n"
"\n"
"See " DRM(Point.failure) " for specifics.\n"
)},

	{"events",
		(PyCFunction) transit_events, METH_NOARGS,
		PyDoc_STR(
"events()\n\n"
":returns: An iterator of integers noting the events that occurred.\n"
":rtype: :py:class:`collections.Iterable`\n"
"\n"
"Constructs and returns an iterator for the events that occurred within the current "
"Traffic cycle.\n"
)},

	{NULL,},
};

static PyMemberDef transit_members[] = {
	{"traffic",
		T_OBJECT, offsetof(struct Transit, traffic), 0,
		PyDoc_STR(
"The " DRC(Traffic) " instance that the Transit has been acquired by.\n"
"`None` if the Transit has not been acquired by a Traffic instance."
)},

	{"link",
		T_OBJECT, offsetof(struct Transit, link), 0,
		PyDoc_STR(
"The 'link-back' to associated objects in the process. ('user data')"
)},

	/* state */
	{"_junction", T_UBYTE, offsetof(struct Transit, junction), READONLY,
		PyDoc_STR("bit map defining the internal and external state of the transit")},
	{"_delta", T_UBYTE, offsetof(struct Transit, delta), READONLY,
		PyDoc_STR("bit map defining the internal state changes")},
	{"_event", T_UBYTE, offsetof(struct Transit, event), READONLY,
		PyDoc_STR("bit map of events that occurred this cycle")},
	{NULL,},
};

static void
transit_dealloc(PyObj self)
{
	Transit t = (Transit) self;

	ptransit(t);
	/*
	 * Traffic instances hold a reference to a Transit until it is
	 * removed from the ring. Transits hold their reference until..now:
	 */

	Py_XDECREF(t->traffic);
	t->traffic = NULL;

	Py_XDECREF(t->link);
	t->link = NULL;
}

static PyObj
transit_get_polarity(PyObj self, void *_)
{
	Transit t = (Transit) self;
	PyObj rob;

	if (Transit_Polarity(t))
		rob = Py_True;
	else
		rob = Py_False;

	Py_INCREF(rob);
	return(rob);
}

static PyObj
transit_get_terminated(PyObj self, void *_)
{
	Transit t = (Transit) self;
	PyObj rob;

	if (Transit_Terminated(t))
		rob = Py_True;
	else
		rob = Py_False;

	Py_INCREF(rob);
	return(rob);
}

static PyGetSetDef transit_getset[] = {
	/* Event introspection. */
	{"polarity", transit_get_polarity, NULL,
		PyDoc_STR("`True` if the transit receives, `False` if it sends.")},
	{"terminated", transit_get_terminated, NULL,
		PyDoc_STR("whether the transit is capable of transfers")},
	{NULL,},
};

const char transit_doc[] =
"Transits are objects that *can* perform transfers until termination.\n"
"In order for events to be processed, the Transit must be acquired by\n"
"a " DRC(Traffic) " instance."
;
TransitPyTypeObject TransitType = {{
	PyVarObject_HEAD_INIT(NULL, 0)
	QPATH("Transit"),					/* tp_name */
	sizeof(struct Transit),			/* tp_basicsize */
	0,										/* tp_itemsize */
	transit_dealloc,					/* tp_dealloc */
	NULL,									/* tp_print */
	NULL,									/* tp_getattr */
	NULL,									/* tp_setattr */
	NULL,									/* tp_compare */
	NULL,									/* tp_repr */
	NULL,									/* tp_as_number */
	NULL,									/* tp_as_sequence */
	NULL,									/* tp_as_mapping */
	NULL,									/* tp_hash */
	NULL,									/* tp_call */
	NULL,									/* tp_str */
	NULL,									/* tp_getattro */
	NULL,									/* tp_setattro */
	NULL,									/* tp_as_buffer */
	Py_TPFLAGS_BASETYPE|
	Py_TPFLAGS_DEFAULT,		  		/* tp_flags */
	transit_doc,						/* tp_doc */
	NULL,									/* tp_traverse */
	NULL,									/* tp_clear */
	NULL,									/* tp_richcompare */
	0,										/* tp_weaklistoffset */
	NULL,									/* tp_iter */
	NULL,									/* tp_iternext */
	transit_methods,					/* tp_methods */
	transit_members,					/* tp_members */
	transit_getset,					/* tp_getset */
	NULL,									/* tp_base */
	NULL,									/* tp_dict */
	NULL,									/* tp_descr_get */
	NULL,									/* tp_descr_set */
	0,										/* tp_dictoffset */
	NULL,									/* tp_init */
	NULL,									/* tp_alloc */
	NULL,									/* tp_new */
},
	NULL,
};

static void
ti_point_kfilter(Transit t, kevent_t kev)
{
	Point p = (Point) t;

	kev->filter = (Transit_Sends(p) ? EVFILT_WRITE : EVFILT_READ);
	kev->ident = Point_GetKPoint(p);
	kev->flags = EV_CLEAR;
	kev->data = 0;
}

static void
ti_point_transform(Transit t, kevent_t kev)
{
	Point p = (Point) t;

	p->ksize = (Py_ssize_t) kev->data;

	/*
	 * Always note when a transfer is *possible*.
	 * The iTransfer must be present in order for an event to be enqueued.
	 */
	if (p->ksize > 0)
		Transit_XQualify(p, teq_transfer);
	else
		Transit_XNQualify(p, teq_transfer);

	if (kev->flags & EV_ERROR)
	{
		/*
		 * Error condition on the file descriptor?
		 */
		Transit_XQualify(t, teq_terminate);
		Point_SetKError(((Point) t), (kerror_t) kev->fflags);
		Point_SetKCall(((Point) t), kc_kevent);
	}
	else if (kev->flags & EV_EOF && kev->data == 0)
	{
		/*
		 * Only xterminate when there's no data to transfer.
		 * This is safe for EVFILT_WRITE as those filters don't "receive" EV_EOF.
		 */
		Transit_XQualify(t, teq_terminate);
		Point_SetKError(((Point) t), 0);
		Point_SetKCall(((Point) t), kc_kevent);
	}
}

static void
ti_point_transfer(Transit t)
{
	return;
}

static void
ti_point_xrelease(Transit t)
{
	Point p = (Point) t;

	if (close(Point_GetKPoint(p)) < 0)
	{
		switch (errno)
		{
			case EPIPE: /* XXX: Apparently, linux throws this. Assume EIO? */
			case EIO:
				/*
				 * some pending write did not complete. note error, but iff
				 * an existing error has not been noted.
				 */
				if (Point_GetKError(p) == 0)
				{
					Point_SetKError(p, EIO);
					Point_SetKCall(p, kc_close);
				}
			break;
		}
	}

	Point_SetKPoint(p, -1);
	Transit_XQualify(p, teq_terminate);
	return;
}

struct TransitInterface PointTIF = {
	ti_point_kfilter,
	ti_point_transform,
	ti_point_transfer,
	ti_transit_terminate,
	ti_point_xrelease,
	ti_transit_irelease,
};

static PyObj
point_failure(PyObj self)
{
	Point p = (Point) self;
	PyObj rob;

	rob = PyTuple_New(2);
	if (rob == NULL)
		return(NULL);

	PyTuple_SET_ITEM(rob, 0, kcall_strob(Point_GetKCall(p)));
	PyTuple_SET_ITEM(rob, 1, PyLong_FromLong(((long) Point_GetKError(p))));

	return(rob);
}

static PyMethodDef point_methods[] = {
	{"failure", (PyCFunction) point_failure,
		METH_NOARGS, PyDoc_STR(
"failure()\n\n"
":returns: A tuple of the function name that returned an error and its errno: "
"(function_name, errno).\n"
":rtype: (:py:class:`str`, :py:class:`int`)\n"
"\n"
"If no failure has occurred, it will return the tuple: ``(None, 0)``.\n"
"In cases where the Point was properly terminated by external forces, the tuple will\n"
"be: ``(\"kevent\", 0)``."
)},

	{NULL,},
};

static PyMemberDef point_members[] = {
	{"_fileno", T_INT, offsetof(struct Point, kpoint), READONLY,
		PyDoc_STR(
"The file descriptor of the Point of instance.\n"
"This attribute will be `-1` *after* termination."
)},
	{"_errno",
		T_INT, offsetof(struct Point, kerror), READONLY,
		PyDoc_STR(
"The error \"number\" associated with the Transit's termination."
)},
	{"_ksize", T_PYSSIZET, offsetof(struct Point, ksize), READONLY,
		PyDoc_STR(
"The number of units available in the associated kernel buffer."
)},
	{NULL,},
};

static PyGetSetDef point_getset[] = {
	{NULL,},
};

static void
point_dealloc(PyObj self)
{
	Point p = (Point) self;

	/*
	 * Traffic instances hold a reference to a point until it is
	 * explicitly closed. At that point, it is detached from the ring,
	 * and Traffic's reference is released.
	 */
	if (Point_GetKPoint(p) != -1)
	{
#if PY_MAJOR_VERSION > 2
		PyErr_WarnFormat(PyExc_ResourceWarning, 0,
			"point transit had valid file descriptor on deallocation");
#endif
		close(Point_GetKPoint(p));
	}

	return(transit_dealloc(self));
}

const char Point_doc[] =
"Points are an abstract Transit whose transfers are facilitated by\n"
"a 'kernel point', a file descriptor on UNIX systems.\n"
;
TransitPyTypeObject PointType = {{
	PyVarObject_HEAD_INIT(NULL, 0)
	QPATH("Point"),			/* tp_name */
	sizeof(struct Point),			/* tp_basicsize */
	0,										/* tp_itemsize */
	point_dealloc,						/* tp_dealloc */
	NULL,									/* tp_print */
	NULL,									/* tp_getattr */
	NULL,									/* tp_setattr */
	NULL,									/* tp_compare */
	NULL,									/* tp_repr */
	NULL,									/* tp_as_number */
	NULL,									/* tp_as_sequence */
	NULL,									/* tp_as_mapping */
	NULL,									/* tp_hash */
	NULL,									/* tp_call */
	NULL,									/* tp_str */
	NULL,									/* tp_getattro */
	NULL,									/* tp_setattro */
	NULL,									/* tp_as_buffer */
	Py_TPFLAGS_BASETYPE|
	Py_TPFLAGS_DEFAULT,		  		/* tp_flags */
	Point_doc,							/* tp_doc */
	NULL,									/* tp_traverse */
	NULL,									/* tp_clear */
	NULL,									/* tp_richcompare */
	0,										/* tp_weaklistoffset */
	NULL,									/* tp_iter */
	NULL,									/* tp_iternext */
	point_methods,						/* tp_methods */
	point_members,						/* tp_members */
	point_getset,						/* tp_getset */
	&TransitType.typ,					/* tp_base */
	NULL,									/* tp_dict */
	NULL,									/* tp_descr_get */
	NULL,									/* tp_descr_set */
	0,										/* tp_dictoffset */
	NULL,									/* tp_init */
	NULL,									/* tp_alloc */
	NULL,									/* tp_new */
},
	&PointTIF
};

/*
 * inet_pton for python string or buffer objects.
 * if it's a unicode object, convert to ascii.
 */
static signed int
inet_from_pystr(void *out, int af, PyObj obj)
{
	unsigned char addrbuf[INET_ADDRSTRLEN];
	Py_buffer view;
	PyObj addr;
	int r;

	if (PyObject_IsInstance(obj, (PyObj) &PyUnicode_Type))
	{
		/* passing this into */
		addr = PyUnicode_AsASCIIString(obj);
		if (addr == NULL)
			return(-1);
	}
	else if (PyObject_CheckBuffer(obj))
	{
		/* unconditional decref later */
		addr = obj;
		Py_INCREF(addr);
	}
	else
	{
		/* not a buffer nor a string, return "bad" type */
		return(1);
	}

	if (PyObject_GetBuffer(addr, &view, 0) != 0)
	{
		/*
		 * it's a buffer, yet getbuffer failed..
		 */
		Py_DECREF(addr);
		return(-1);
	}

	/* >= because addrbuf has space for terminator */
	if (view.len >= sizeof(addrbuf))
	{
		PyErr_SetString(PyExc_ValueError,
			"string given as an IPv4 address is too long");
		Py_DECREF(addr);
		return(-1);
	}

	memcpy(addrbuf, view.buf, view.len);
	addrbuf[view.len] = '\0';
	r = inet_pton(af, (char *) addrbuf, out);
	Py_DECREF(addr);

	return(0);
}

/*
 * Perform the buffer operation against the kpoint
 */
static kerror_t
perform(kp_buffer_op_t op, kpoint_t kp, void *buf, int amount, int *xferout)
{
	int xferd;

dosyscall:
	xferd = op(kp, buf, amount);
	if (xferd < 0)
	{
		switch (errno)
		{
			case EINTR:
				/* interrupted. try again. */
				goto dosyscall;
			break;

			case EWOULDBLOCK:
			case ETIMEDOUT:
				/*
				 * These are passed on as fatal errors as it means
				 * someone wrote to the file descriptor making our
				 * state inconsistent with the kernel's.
				 * That's a problem.
				 */
			default:
				/* fatal */
				*xferout = 0;
				return(errno);
			break;
		}
	}

	*xferout = xferd;
	return(0);
}

static void
ti_octets_transfer(Transit t)
{
	Octets o = (Octets) t;
	int err = 0, remaining, xfer = 0;
	void *pos;

	/*
	 * Get position and remainder in current working buffer.
	 */
	pos = (void *) (((intptr_t) (o->view).buf) + o->position);
	remaining = (int) ((o->view).len - o->position);

	/*
	 * Perform the write, as much as possible.
	 */
	err = perform(Octets_Mechanism(o), Point_GetKPoint(o), pos, remaining, &xfer);

#ifdef PRINTF_DEBUGGING
	printf("XFER %s: %d/%d\n", Transit_Sends(o) ? "OUT" : "IN", xfer, remaining);
#endif

	if (err != 0)
	{
		/*
		 * Fatal. perform() handles EINTR cases, and
		 * the call spec should avoid any non-fatal error conditions.
		 */
		Point_SetKError(o, err);
		Point_SetKCall(o, Transit_Sends(o) ? kc_write : kc_read);
		Transit_IQualify(o, teq_terminate);
		Transit_XQualify(o, teq_terminate);
	}
	else
	{
		/* Success! Update the buffer position and PROFIT! */
		o->last_position = o->position;
		o->position += xfer;
		o->ksize -= xfer;

		if (o->position == o->view.len)
		{
			/*
			 * Exhausted the resource. Update junction state.
			 */
			Transit_INQualify(o, teq_transfer);
			Transit_NoteEvent(o, tev_exhaust);
		}

		Transit_NoteEvent(o, tev_transfer);
	}

	if (!(o->ksize > 0))
	{
		/* exhausted kernel resource; wait for kevent now */
		Transit_XNQualify(o, teq_transfer);
	}
}

static PyObj
octets_acquire(PyObj self, PyObj resource)
{
	Octets o = (Octets) self;
	char needs_writable;

	if (!transit_can_acquire((Transit) o))
		return(NULL);

	Transit_GetInterface(o)->ti_irelease((Transit) o);

	o->resource = resource;
	Py_INCREF(o->resource);

	if (PyObject_GetBuffer(o->resource, &(o->view), Transit_Receives(o) ? PyBUF_WRITABLE : 0))
	{
		Py_DECREF(o->resource);
		o->resource = NULL;
		return(NULL);
	}

	o->last_position = o->position = 0;

	transit_acquired((Transit) o);
	Py_INCREF(self);
	return(self);
}

static PyObj
octets_transfer(PyObj self)
{
	Octets o = (Octets) self;
	return(_PySlice_FromIndices(o->last_position, o->position));
}

static PyObj
socket_identity(Point p, unsigned char peer)
{
	PyObj rob = NULL;
	PyObj host, port;
	struct sockaddr_storage addr;
	socklen_t addrlen = sizeof(addr);
	int r;

	if (peer)
		r = getpeername(Point_GetKPoint(p), (struct sockaddr *) &addr, &addrlen);
	else
		r = getsockname(Point_GetKPoint(p), (struct sockaddr *) &addr, &addrlen);

	if (r != 0)
	{
		PyErr_SetFromErrno(PyExc_OSError);
		return(NULL);
	}

	switch (addr.ss_family)
	{
		case AF_INET:
		{
			struct sockaddr_in *addr_af_inet;
			char dst[INET_ADDRSTRLEN];
			addr_af_inet = (struct sockaddr_in *) &addr;

			if (inet_ntop(AF_INET, &(addr_af_inet->sin_addr), dst, sizeof(dst)) == NULL)
			{
				PyErr_SetFromErrno(PyExc_OSError);
				return(NULL);
			}

			host = PyUnicode_FromString(dst);
			if (host != NULL)
			{
				port = PyLong_FromLong((long) ntohs(addr_af_inet->sin_port));
				if (port == NULL)
				{
					Py_DECREF(host);
					return(NULL);
				}
			}
			else
				return(NULL);
		}
		break;

		case AF_INET6:
		{
			struct sockaddr_in6 *addr_af_inet6;
			char dst[INET6_ADDRSTRLEN];
			addr_af_inet6 = (struct sockaddr_in6 *) &addr;

			if (inet_ntop(AF_INET6, &(addr_af_inet6->sin6_addr), dst, sizeof(dst)) == NULL)
			{
				PyErr_SetFromErrno(PyExc_OSError);
				return(NULL);
			}

			host = PyUnicode_FromString(dst);
			if (host != NULL)
			{
				port = PyLong_FromLong((long) ntohs(addr_af_inet6->sin6_port));
				if (port == NULL)
				{
					Py_DECREF(host);
					return(NULL);
				}
			}
			else
				return(NULL);
		}
		break;

		case AF_LOCAL:
		{
			struct sockaddr_un *addr_af_local;
			addr_af_local = (struct sockaddr_un *) &addr;

			host = PyBytes_FromString(addr_af_local->sun_path);
			port = Py_None;
			Py_INCREF(port);
		}
		break;

		default:
			PyErr_SetString(PyExc_RuntimeError, "unknown address family");
			return(NULL);
		break;
	}

	rob = PyTuple_New(2);
	if (rob != NULL)
	{
		PyTuple_SET_ITEM(rob, 0, host);
		PyTuple_SET_ITEM(rob, 1, port);
	}
	else
	{
		Py_DECREF(host);
		Py_DECREF(port);
	}

	return(rob);
}

static PyObj
octets_identity(PyObj self)
{
	Octets o = (Octets) self;

	if (Octets_IsSocket(o))
		return(socket_identity((Point) o, Transit_Sends(o)));
	else
		return(PyLong_FromLong((long) Point_GetKPoint(o)));
}

/*
 * Create a pair of Octets and put them in a tuple to be returned.
 */
static PyObj
streampair(PyObj subtype)
{
	PyObj rob;
	Octets input, output;

	rob = PyTuple_New(2);
	if (rob == NULL)
		return(NULL);

	input = (Octets) PyAllocate(subtype);
	if (input == NULL)
	{
		Py_DECREF(rob);
		return(NULL);
	}
	PyTuple_SET_ITEM(rob, 0, (PyObj) input);

	output = (Octets) PyAllocate(subtype);
	if (output == NULL)
	{
		Py_DECREF(rob);
		return(NULL);
	}
	PyTuple_SET_ITEM(rob, 1, (PyObj) output);

	return(rob);
}

/*
 * connect_point - connect an arbitrary point
 */
static PyObj
connect_point(PyObj subtype, int fd, char polarity)
{
	blksize_t blksize;
	kerror_t ke = 0;
	kcall_t kc = kc_none;
	kpoint_t kp = (kpoint_t) fd;
	Octets o;

	o = (Octets) PyAllocate(subtype);
	if (o == NULL)
		return(NULL);

	if (init_kpoint(kp))
		ke = errno;

	if (polarity)
		INIT_READ_OCTETS(o, kp, kc, ke);
	else
		INIT_WRITE_OCTETS(o, kp, kc, ke);

	Octets_SetKType(o, point_ktype(fd, &blksize));

	return((PyObj) o);
}

/*
 * Octets pair from socket file descriptor.
 */
static PyObj
octets_connect_input_point(PyObj subtype, PyObj args)
{
	int fd;

	if (!PyArg_ParseTuple(args, "i", &fd))
		return(NULL);

	return(connect_point(subtype, fd, 1));
}

/*
 * Octets pair from socket file descriptor.
 */
static PyObj
octets_connect_output_point(PyObj subtype, PyObj args)
{
	int fd;

	if (!PyArg_ParseTuple(args, "i", &fd))
		return(NULL);

	return(connect_point(subtype, fd, 0));
}

static PyObj
connect_socket(PyObj subtype, int socketfd)
{
	kerror_t ke = 0;
	kcall_t kc = kc_none;
	PyObj rob;

	rob = streampair(subtype);
	if (rob == NULL)
		return(NULL);

	if (init_socket(socketfd))
		ke = errno;

	INIT_PAIR_TUPLE(rob, socketfd, socketfd, kc, ke);
	return(rob);
}

/*
 * Octets pair from socket file descriptor.
 */
static PyObj
octets_connect_stream_socket(PyObj subtype, PyObj args)
{
	int fd;

	if (!PyArg_ParseTuple(args, "i", &fd))
		return(NULL);

	return(connect_socket(subtype, fd));
}

/*
 * traffic.libkqueue.Octets.connect_socketpair()
 * Construct a pair of Octets using socketpair().
 */
static PyObj
octets_connect_socketpair(PyObj subtype)
{
	int fdv[2] = {-1,-1};
	PyObj rob, p1, p2;
	kcall_t kc = kc_none;
	kerror_t ke = 0;

	p1 = streampair(subtype);
	if (p1 == NULL)
		return(NULL);

	p2 = streampair(subtype);
	if (p2 == NULL)
	{
		Py_DECREF(p1);
		return(NULL);
	}

	rob = PyTuple_New(2);
	if (rob == NULL)
	{
		Py_DECREF(p1);
		Py_DECREF(p2);
		return(NULL);
	}

	PyTuple_SET_ITEM(rob, 0, p1);
	PyTuple_SET_ITEM(rob, 1, p2);

	if (!socketpair(AF_LOCAL, SOCK_STREAM, 0, fdv))
	{
		init_socket(fdv[0]);
		init_socket(fdv[1]);
	}
	else
	{
		kc = kc_socketpair;
		ke = errno;
	}

	INIT_PAIR_TUPLE(p1, fdv[0], fdv[0], 0, 0);
	INIT_PAIR_TUPLE(p2, fdv[1], fdv[1], 0, 0);

	return(rob);
}

/*
 * connect_ip4 - Make a IPv4 Connection.
 */
static PyObj
connect_ip4(int socktype, PyObj subtype,
	PyObj host_addr_ob, short host_port,
	PyObj bind_addr_ob, short bind_port)
{
	struct sockaddr_in host_addr = {0,}, bind_addr = {0,};
	PyObj rob;
	int r;
	kpoint_t kp = -1;
	kerror_t ke = 0;
	kcall_t kc = kc_none;

	r = inet_from_pystr(&(host_addr.sin_addr.s_addr), AF_INET, host_addr_ob);
	if (r < 0)
		return(NULL);
	if (r > 0)
	{
		/* not processed / not a string or buffer */
		if (inet4_from_pyint(&(host_addr.sin_addr.s_addr), host_addr_ob))
			return(NULL);
		if (PyErr_Occurred())
			return(NULL);
	}
	host_addr.sin_family = AF_INET;
	SETADDRLEN(host_addr.sin_len, sizeof(host_addr));
	host_addr.sin_port = htons(host_port);

	/* bind address is optional */
	if (bind_addr_ob != NULL)
	{
		bind_addr.sin_family = AF_INET;
		SETADDRLEN(bind_addr.sin_len, sizeof(bind_addr));

		r = inet_from_pystr(&(bind_addr.sin_addr.s_addr), AF_INET, bind_addr_ob);
		if (r < 0)
		{
			/* Invalid address. */
			return(NULL);
		}
		if (r > 0)
		{
			/* not processed / not a string or buffer */
			if (inet4_from_pyint(&(bind_addr.sin_addr.s_addr), bind_addr_ob))
				return(NULL);
		}

		bind_addr.sin_port = htons(bind_port);
	}

	rob = streampair(subtype);
	if (rob == NULL)
		return(NULL);

	Py_BEGIN_ALLOW_THREADS

	kp = socket(PF_INET, socktype, IPPROTO_IP);
	if (kp < 0)
	{
		kc = kc_socket;
		ke = errno;
	}
	else
	{
		init_socket((kpoint_t) kp);

		/*
		 * Bind to requested interface and port.
		 */
		if (bind_addr_ob != NULL)
		{
			if (!bind(kp, (struct sockaddr *) &bind_addr, sizeof(bind_addr)))
			{
				kc = kc_bind;
				ke = errno;
				goto skip_connect;
			}
		}

		if (!connect(kp, (struct sockaddr *) &host_addr, sizeof(host_addr)))
		{
			switch (errno)
			{
				case EINPROGRESS:
				case EISCONN:
					/*
					 * It's connecting or ... it's already connected?
					 * Better be. Non-blocking.
					 *
					 * These are working as expected "errors".
					 */
				break;
				default:
					kc = kc_connect;
					ke = errno;
				break;
			}
		}
skip_connect: ; /* used in bind() failures */
	}

	Py_END_ALLOW_THREADS

	INIT_PAIR_TUPLE(rob, kp, kp, kc, ke);
	return(rob);
}

/*
 * connect_ip6 - Make a IPv6 Connection.
 */
static PyObj
connect_ip6(int socktype, PyObj subtype,
	PyObj host_addr_ob, short host_port,
	PyObj bind_addr_ob, short bind_port)
{
	struct sockaddr_in6 host_addr = {0,}, bind_addr = {0,};
	PyObj rob;
	int r;
	kpoint_t kp = -1;
	kerror_t ke = 0;
	kcall_t kc = kc_none;

	r = inet_from_pystr(&(host_addr.sin6_addr), AF_INET6, host_addr_ob);
	if (r < 0)
		return(NULL);
	if (r > 0)
	{
		/* not processed / not a string or buffer */
		if (inet6_from_pyint(&(host_addr.sin6_addr), host_addr_ob))
			return(NULL);
		if (PyErr_Occurred())
			return(NULL);
	}
	host_addr.sin6_family = AF_INET6;
	SETADDRLEN(host_addr.sin6_len, sizeof(host_addr));
	host_addr.sin6_port = htons(host_port);

	/* bind address is optional */
	if (bind_addr_ob != NULL)
	{
		bind_addr.sin6_family = AF_INET6;
		SETADDRLEN(bind_addr.sin6_len, sizeof(bind_addr));

		r = inet_from_pystr(&(bind_addr.sin6_addr), AF_INET6, bind_addr_ob);
		if (r < 0)
		{
			/* Invalid address. */
			return(NULL);
		}
		if (r > 0)
		{
			/* not processed / not a string or buffer */
			if (inet6_from_pyint(&(bind_addr.sin6_addr), bind_addr_ob))
				return(NULL);
		}

		bind_addr.sin6_port = htons(bind_port);
	}

	rob = streampair(subtype);
	if (rob == NULL)
		return(NULL);

	Py_BEGIN_ALLOW_THREADS

	kp = socket(PF_INET6, socktype, IPPROTO_IP);
	if (kp < 0)
	{
		kc = kc_socket;
		ke = errno;
	}
	else
	{
		init_socket(kp);

		/*
		 * Bind to requested interface and port.
		 */
		if (bind_addr_ob != NULL)
		{
			if (!bind(kp, (struct sockaddr *) &bind_addr, sizeof(bind_addr)))
			{
				kc = kc_bind;
				ke = errno;
				goto skip_connect;
			}
		}

		if (!connect(kp, (struct sockaddr *) &host_addr, sizeof(host_addr)))
		{
			switch (errno)
			{
				case EINPROGRESS:
				case EISCONN:
					/*
					 * It's connecting or it's connected.
					 */
				break;
				default:
					kc = kc_connect;
					ke = errno;
				break;
			}
		}
skip_connect: ; /* used in bind() failures */
	}

	Py_END_ALLOW_THREADS

	INIT_PAIR_TUPLE(rob, kp, kp, kc, ke);
	return(rob);
}

/*
 * connect_local - Make a local file system connection.
 */
static PyObj
connect_local(int socktype, PyObj subtype, const char *path, Py_ssize_t pathsize)
{
	struct sockaddr_un host_addr;
	PyObj rob;
	int r;
	kpoint_t kp = -1;
	kerror_t ke = 0;
	kcall_t kc = kc_none;

	if (pathsize > sizeof(host_addr.sun_path))
	{
		/* XXX: better exception (argument error, EINVAL?) */
		PyErr_SetString(PyExc_ValueError, "local filesystem path too long");
		return(NULL);
	}

	rob = streampair(subtype);
	if (rob == NULL)
		return(NULL);

	Py_BEGIN_ALLOW_THREADS

	host_addr.sun_family = PF_LOCAL;
	strncpy((char *) &(host_addr.sun_path), path, sizeof(host_addr.sun_path));

	kp = socket(PF_LOCAL, socktype, IPPROTO_IP);
	if (kp < 0)
	{
		kc = kc_socket;
		ke = errno;
	}
	else
	{
		init_socket(kp);

		if (!connect(kp, (struct sockaddr *) &host_addr, sizeof(host_addr)))
		{
			switch (errno)
			{
				case EINPROGRESS:
				case EISCONN:
					/*
					 * It's connecting or it's connected. That's okay.
					 */
				break;
				default:
					kc = kc_connect;
					ke = errno;
				break;
			}
		}
	}
	Py_END_ALLOW_THREADS

	INIT_PAIR_TUPLE(rob, kp, kp, kc, ke);
	return(rob);
}

static PyObj
octets_connect_stream_local(PyObj subtype, PyObj args, PyObj kw)
{
	static char *kwlist[] = {"path", NULL};
	const char *path;
	Py_ssize_t size;

	if (!PyArg_ParseTupleAndKeywords(args, kw, "y#", kwlist, &path, &size))
		return(NULL);

	return(connect_local(SOCK_STREAM, subtype, path, size));
}

static PyObj
octets_connect_stream_ip4(PyObj subtype, PyObj args, PyObj kw)
{
	static char *kwlist[] = {"host", "interface", NULL};
	PyObj host_addr_ob, bind_addr_ob = NULL;
	unsigned short host_port, bind_port;

	if (!PyArg_ParseTupleAndKeywords(args, kw, "(OH)|(OH)", kwlist,
			&host_addr_ob, &host_port,
			&bind_addr_ob, &bind_port))
		return(NULL);

	return(connect_ip4(SOCK_STREAM, subtype, host_addr_ob, host_port, bind_addr_ob, bind_port));
}

static PyObj
octets_connect_datagram_ip4(PyObj subtype, PyObj args, PyObj kw)
{
	static char *kwlist[] = {"host", "interface", NULL};
	PyObj host_addr_ob, bind_addr_ob = NULL;
	unsigned short host_port, bind_port;

	if (!PyArg_ParseTupleAndKeywords(args, kw, "(OH)|(OH)", kwlist,
			&host_addr_ob, &host_port,
			&bind_addr_ob, &bind_port))
		return(NULL);

	return(connect_ip4(SOCK_DGRAM, subtype, host_addr_ob, host_port, bind_addr_ob, bind_port));
}

static PyObj
octets_connect_stream_ip6(PyObj subtype, PyObj args, PyObj kw)
{
	static char *kwlist[] = {"host", "interface", NULL};
	PyObj host_addr_ob, bind_addr_ob = NULL;
	unsigned short host_port, bind_port;

	if (!PyArg_ParseTupleAndKeywords(args, kw, "(OH)|(OH)", kwlist,
			&host_addr_ob, &host_port,
			&bind_addr_ob, &bind_port))
		return(NULL);

	return(connect_ip6(SOCK_STREAM, subtype, host_addr_ob, host_port, bind_addr_ob, bind_port));
}

static PyObj
octets_connect_datagram_ip6(PyObj subtype, PyObj args, PyObj kw)
{
	static char *kwlist[] = {"host", "interface", NULL};
	PyObj host_addr_ob, bind_addr_ob = NULL;
	unsigned short host_port, bind_port;

	if (!PyArg_ParseTupleAndKeywords(args, kw, "(OH)|(OH)", kwlist,
			&host_addr_ob, &host_port,
			&bind_addr_ob, &bind_port))
		return(NULL);

	return(connect_ip6(SOCK_DGRAM, subtype, host_addr_ob, host_port, bind_addr_ob, bind_port));
}

static PyObj
octets_connect_pipe(PyObj subtype)
{
	PyObj rob;
	int fds[2];
	kcall_t kc = kc_none;
	kerror_t ke = 0;

	rob = streampair(subtype);
	if (rob == NULL)
		return(NULL);

	Py_BEGIN_ALLOW_THREADS

	if (pipe(fds) < 0)
	{
		kc = kc_pipe;
		ke = errno;
	}

	Py_END_ALLOW_THREADS

	INIT_PAIR_TUPLE(rob, fds[0], fds[1], kc, ke);
	return(rob);
}

static PyObj
octets_connect_stdio(PyObj subtype)
{
	PyObj rob;

	rob = streampair(subtype);
	if (rob == NULL)
		return(NULL);

	INIT_PAIR_TUPLE(rob, 0, 1, kc_none, 0);

	return(rob);
}

static PyObj
octets_connect_stdie(PyObj subtype)
{
	PyObj rob;
	int in = -1, out = -1;
	kerror_t ke = 0;

	in = dup(0);
	if (in < 0)
		ke = errno;
	else
	{
		out = dup(2);
		if (out < 0)
			ke = errno;
	}

	rob = streampair(subtype);
	if (rob == NULL)
	{
		close(in);
		close(out);
		return(NULL);
	}

	INIT_PAIR_TUPLE(rob, in, out, kc_dup, ke);

	return(rob);
}

static PyObj
octets_connect_stderr(PyObj subtype)
{
	PyObj rob;
	Octets o;

	rob = PyAllocate(subtype);
	if (rob == NULL)
		return(NULL);

	o = (Octets) rob;
	INIT_WRITE_OCTETS(o, 2, kc_none, 0);

	return(rob);
}

static PyObj
connect_file(PyObj subtype, PyObj args, PyObj kw, int oflags, unsigned char sends)
{
	static char *kwlist[] = {"path", "start", NULL};
	int fd, whence = 0;
	kerror_t ke = 0;
	kcall_t kc = kc_none;
	char *path;
	long long llstart = 0;
	off_t start = 0;
	PyObj rob;
	Octets o;

	if (!PyArg_ParseTupleAndKeywords(args, kw, "y|L", kwlist, &path, &llstart))
		return(NULL);

	if (llstart < 0)
	{
		whence = SEEK_END;
		start = (off_t) -llstart;
	}
	else
		start = (off_t) llstart;

	rob = PyAllocate(subtype);
	if (rob == NULL)
		return(NULL);
	o = (Octets) rob;
	o->ktype = octets_file;

	fd = open(path, oflags);
	if (fd < 0)
	{
		ke = (kerror_t) errno;
		kc = kc_open;
	}

	if (ke == 0 && start != 0)
	{
		int r;
		r = lseek(fd, start, whence);

		if (r < 0)
		{
			ke = (kerror_t) errno;
			kc = kc_lseek;
		}
	}

	if (sends)
		INIT_WRITE_OCTETS(o, fd, kc, ke);
	else
		INIT_READ_OCTETS(o, fd, kc, ke);

	return(rob);
}

static PyObj
octets_connect_file(PyObj subtype, PyObj args, PyObj kw)
{
	return(connect_file(subtype, args, kw, O_NONBLOCK|O_RDONLY, 0));
}

static PyObj
octets_connect_file_overwrite(PyObj subtype, PyObj args, PyObj kw)
{
	return(connect_file(subtype, args, kw, O_NONBLOCK|O_WRONLY|O_CREAT, 1));
}

static PyObj
octets_connect_file_append(PyObj subtype, PyObj args, PyObj kw)
{
	return(connect_file(subtype, args, kw, O_NONBLOCK|O_WRONLY|O_APPEND|O_CREAT, 1));
}

static PyObj
octets_rallocate(PyObj subtype, PyObj size)
{
	PyObj args, rob;

	args = PyTuple_New(1);
	if (args == NULL)
		return(NULL);

	PyTuple_SET_ITEM(args, 0, PyByteArray_FromObject(size));
	if (PyTuple_GET_ITEM(args, 0) == NULL)
	{
		Py_DECREF(args);
		return(NULL);
	}

	rob = PyObject_CallObject(((PyObj) &PyMemoryView_Type), args);
	Py_DECREF(args);

	return(rob);
}

/* METH_O, METH_VARARGS, METH_KEYWORDS, METH_NOARGS */
static PyMethodDef octets_methods[] = {
/* resource allocation */
	{"rallocate",
		(PyCFunction) octets_rallocate,
		METH_O|METH_CLASS,
		PyDoc_STR(
"rallocate(size)\n\n"
":param size: Number of bytes that can be written in the resource.\n"
":type size: :py:class:`int`\n"
":returns: A new mutable resource.\n"
":rtype: :py:class:`memoryview`(:py:class:`bytearray`)\n"
"\n"
"Create a mutable resource capable of being written into by a Octets instance.\n"
)},

/* creation */
	{"connect_input_point", (PyCFunction) octets_connect_input_point,
		METH_VARARGS|METH_CLASS,
		PyDoc_STR(
"connect_input_point(fileno)\n\n"
":param fileno: The file descriptor referring to the kernel resource.\n"
":type fileno: :py:class:`int`\n"
":returns: octets: A single Octets instance that performs reads.\n"
":rtype: " DRC(Octets) "\n"
"\n"
"Create and return an Octets instance for reading from the given file descriptor.\n"
"If the file descriptor is a socket, use " DRM(Octets.connect_stream_socket) "\n"
"**The given file descriptor's options are modified to perform non-blocking I/O\n"
"and disable SIGPIPE on supporting platforms.**"
)},

	{"connect_output_point", (PyCFunction) octets_connect_output_point,
		METH_VARARGS|METH_CLASS,
		PyDoc_STR(
"connect_output_point(fileno)\n\n"
":param fileno: The file descriptor referring to the kernel resource.\n"
":type fileno: :py:class:`int`\n"
":returns: octets: A single Octets instance that performs writes.\n"
":rtype: " DRC(Octets) "\n"
"\n"
"Create and return an Octets instance for writing to the given file descriptor.\n"
"If the file descriptor is a socket, use " DRM(Octets.connect_stream_socket) "\n"
"**The given file descriptor's options are modified to perform non-blocking I/O\n"
"and disable SIGPIPE on supporting platforms.**"
)},

	{"connect_stream_socket", (PyCFunction) octets_connect_stream_socket,
		METH_VARARGS|METH_CLASS,
		PyDoc_STR(
"connect_stream_socket(fileno)\n\n"
":param fileno: The file descriptor associated with the socket to be used.\n"
":type fileno: :py:class:`int`\n"
":returns: (input, output): A pair of Octets.\n"
":rtype: (" DRC(Octets) ", " DRC(Octets) ")\n"
"\n"
"Create and return a pair--`(input, output)`--of Octets instances reading from\n"
"and writing to a SOCK_STREAM socket identified by the first argument.\n"
"\n"
"**The given socket's options are modified to perform non-blocking I/O\n"
"and disable SIGPIPE on supporting platforms.**"
)},

	{"connect_stream_ip4", (PyCFunction) octets_connect_stream_ip4,
		METH_VARARGS|METH_KEYWORDS|METH_CLASS,
		PyDoc_STR(
"connect_stream_ip4(host, interface = None)\n\n"
":param host: The `(ip4_address, port)` pair to connect to.\n"
":type host: (:py:class:`str`, :py:class:`int`)\n"
":param interface: The `(local_interface, port)` pair to bind to.\n"
":type interface: (:py:class:`str`, :py:class:`int`)\n"
":returns: (input, output): A pair of Octets.\n"
":rtype: (" DRC(Octets) ", " DRC(Octets) ")\n"
"\n"
"Create and return a pair of Octets reading from\n"
"and writing to a new SOCK_STREAM/IPv4 connection established using\n"
":manpage:`socket(2)` and :manpage:`connect(2)`.\n"
)},

	{"connect_stream_ip6", (PyCFunction) octets_connect_stream_ip6,
		METH_VARARGS|METH_KEYWORDS|METH_CLASS,
		PyDoc_STR(
"connect_stream_ip6(host, interface = None)\n\n"
":param host: The `(ip6_address, port)` pair to connect to.\n"
":type host: (:py:class:`str`, :py:class:`int`)\n"
":param interface: The `(local_interface, port)` pair to bind to.\n"
":type interface: (:py:class:`str`, :py:class:`int`)\n"
":returns: (input, output)\n"
":rtype: (" DRC(Octets) ", " DRC(Octets) ")\n"
"\n"
"Create and return a pair of Octets reading from\n"
"and writing to a new SOCK_STREAM/IPv6 connection established using\n"
":manpage:`connect(2)`.\n"
)},

	{"connect_stream_local", (PyCFunction) octets_connect_stream_local,
		METH_VARARGS|METH_KEYWORDS|METH_CLASS,
		PyDoc_STR(
"connect_stream_local(path)\n\n"
":param path: The file system path to the local socket.\n"
":type path: :py:class:`bytes`\n"
":returns: (input, output): A pair of Octets.\n"
":rtype: (" DRC(Octets) ", " DRC(Octets) ")\n"
"\n"
"Create and return a pair--`(input, output)`--of Octets instances reading from\n"
"and writing to a new SOCK_STREAM/AF_LOCAL connection.\n"
)},

	{"connect_datagram_ip4", (PyCFunction) octets_connect_datagram_ip4,
		METH_VARARGS|METH_KEYWORDS|METH_CLASS,
		PyDoc_STR(
"connect_datagram_ip4(host, interface = None)\n\n"
":param host: The `(ip4_address, port)` pair to send datagrams to.\n"
":type host: (:py:class:`str`, :py:class:`int`)\n"
":param interface: The `(local_interface, port)` pair to bind to.\n"
":type interface: (:py:class:`str`, :py:class:`int`)\n"
":returns: (input, output): A pair of Octets.\n"
":rtype: tuple((" DRC(Octets) ", " DRC(Octets) "))\n"
"\n"
"Create and return a pair--`(input, output)`--of Octets instances reading from\n"
"and writing to a new SOCK_DATAGRAM/IPv4 connection.\n"
)},

	{"connect_datagram_ip6", (PyCFunction) octets_connect_datagram_ip6,
		METH_VARARGS|METH_KEYWORDS|METH_CLASS,
		PyDoc_STR(
"connect_datagram_ip6(host, interface = None)\n\n"
":param host: The `(ip6_address, port)` pair to send datagrams to.\n"
":type host: (:py:class:`str`, :py:class:`int`)\n"
":param interface: The `(local_interface, port)` pair to bind to.\n"
":type interface: (:py:class:`str`, :py:class:`int`)\n"
":returns: (input, output): A pair of Octets.\n"
":rtype: (" DRC(Octets) ", " DRC(Octets) ")\n"
"\n"
"Create and return a pair--`(input, output)`--of Octets instances reading from\n"
"and writing to a new SOCK_DATAGRAM/IPv6 connection.\n"
)},

	{"connect_stream_socketpair", (PyCFunction) octets_connect_socketpair,
		METH_NOARGS|METH_CLASS,
		PyDoc_STR(
"connect_stream_socketpair()\n\n"
":returns: ((input, output), (input, output)).\n"
":rtype: ((" DRC(Octets) ", " DRC(Octets) "), (" DRC(Octets) ", " DRC(Octets) "))\n"
"\n"
"Construct and return two pairs of Octets connected to each other.\n"
"The sockets are created using :manpage:`socketpair(2)`"
)},

	{"connect_pipe", (PyCFunction) octets_connect_pipe,
		METH_NOARGS|METH_CLASS,
		PyDoc_STR(
":returns: (input, output)\n"
":rtype: (" DRC(Octets) ", " DRC(Octets) ")\n"
"\n"
"Return a pair of a Octets connected to each other.\n"
"The former writing into the latter.\n"
"The file descriptors are created using :manpage:`pipe(2)`"
)},

	{"connect_stdio", (PyCFunction) octets_connect_stdio,
		METH_NOARGS|METH_CLASS,
		PyDoc_STR(
":returns: (input, output)\n"
":rtype: (" DRC(Octets) ", " DRC(Octets) ")\n"
"\n"
"Return a pair of Octets from standard input and standard output."
)},

	{"connect_stdie", (PyCFunction) octets_connect_stdio,
		METH_NOARGS|METH_CLASS,
		PyDoc_STR(
":returns: (input, output)\n"
":rtype: (" DRC(Octets) ", " DRC(Octets) ")\n"
"\n"
"Return a pair of Octets using standard input and standard error."
)},

	{"connect_stderr", (PyCFunction) octets_connect_stderr,
		METH_NOARGS|METH_CLASS,
		PyDoc_STR(
":returns: A new Octets instance.\n"
":rtype: " DRC(Octets) "\n"
"\n"
"Create an output Octets to standard error."
)},

	{"connect_file",
		(PyCFunction) octets_connect_file,
		METH_VARARGS|METH_KEYWORDS|METH_CLASS,
		PyDoc_STR(
"connect_file(path)\n\n"
":param path: The path to the file.\n"
":type path: :py:class:`bytes`\n"
":returns: A new input Octets instance.\n"
":rtype: " DRC(Octets) "\n"
"\n"
".. warning:: Broken\n\n"
"Create an input Octets reading from the given file at the given path."
)},

	{"connect_file_overwrite",
		(PyCFunction) octets_connect_file_overwrite,
		METH_VARARGS|METH_KEYWORDS|METH_CLASS,
		PyDoc_STR(
"connect_file_overwrite(path)\n\n"
":param path: The path to the file.\n"
":type path: :py:class:`bytes`\n"
":returns: A new output Octets instance.\n"
":rtype: " DRC(Octets) "\n"
"\n"
".. warning:: Broken\n\n"
"Create an output Octets overwriting to the file at the given path."
)},

	{"connect_file_append",
		(PyCFunction) octets_connect_file_append,
		METH_VARARGS|METH_KEYWORDS|METH_CLASS,
		PyDoc_STR(
"connect_file_append(path)\n\n"
":param path: The path to the file.\n"
":type path: :py:class:`bytes`\n"
":returns: A new output Octets instance.\n"
":rtype: " DRC(Octets) "\n"
"\n"
".. warning:: Broken\n\n"
"Create an output Octets appending to the file at the given path.\n"
)},

	{"acquire", (PyCFunction) octets_acquire, METH_O,
		PyDoc_STR(
"acquire(buffer)\n\n"
":param buffer: The buffer to write data from or read data into.\n"
":type buffer: `buffer`\n"
":returns: The Octets instance acquiring the resource. (Method Chaining)\n"
":rtype: " DRC(Octets) "\n"
"\n"
"Set the buffer to facilitate transfers with. If the Octets instance receives,\n"
"the buffer *must* be mutable. Usually, the type of mutable buffer used is the built-in\n"
":py:class:`bytearray`.\n"
)},

	{"transfer", (PyCFunction) octets_transfer, METH_NOARGS,
		PyDoc_STR(
"transfer()\n\n"
":returns: A slice specifying the portion of the resource that was transferred.\n"
":rtype: :py:class:`slice`\n"
"\n"
"For input Octets, the slice is often directly applied to the resource in order to\n"
"send the newly received data to a higher-level: ``t.resource[t.transfer()]``.\n\n"
"For output Octets, the slice is often used to identify the number of bytes transferred: ``s=t.transfer(); size=s.stop-s.start``."
)},

/* representation information */
	{"identity", (PyCFunction) octets_identity, METH_NOARGS,
		PyDoc_STR(
"identity()\n\n"
":returns: File descriptor or interface-port pair.\n"
":rtype: [:py:class:`int` or (:py:class:`str`, :py:class:`int`)]\n"
"\n"
"Get the represented endpoint's identity (getpeername or getsockname)."
)},

	{NULL,},
};

static PyMemberDef octets_members[] = {
	{"resource", T_OBJECT, offsetof(struct Octets, resource), READONLY,
		PyDoc_STR(
"The object whose buffer was acquired, " DRM(Octets.acquire) ", as the Transit's transfer resource.\n\n:py:obj:`None` if there is no resource."
)},
	{NULL,},
};

static PyGetSetDef octets_getset[] = {
	{NULL,},
};

static void
ti_octets_terminate(Transit t)
{
	Octets o = (Octets) t;

	if (Octets_IsSocket(o))
	{
		/*
		 * Terminate the representation iff the kevent hasn't indicated it
		 * already.
		 */
		if (!Transit_XQualified(o, teq_terminate))
		{
			shutdown(Point_GetKPoint(o), Transit_Receives(o) ? SHUT_RD : SHUT_WR);
			Transit_XQualify(o, teq_terminate); /* SIGNALS THE OPPOSITE THAT THEIR OPPOSITE IS TERM'd */
		}
		else
		{
			/*
			 * Terminate called, but Xterminate was already set.
			 * This is termination completion, so Iterminate `o`.
			 */
			Transit_IQualify(o, teq_terminate);
		}
	}
}

static void
ti_octets_xrelease(Transit t)
{
	Octets o = (Octets) t;

	if (Octets_IsSocket(o))
	{
		Octets oo = Octets_GetOpposite(o);
		/*
		 * If the opposite is NULL'd, then it can close(),
		 * otherwise, NULL out the opposite's opposite,
		 * signalling it to perform the close().
		 */
		if (oo != NULL && !Transit_XQualified(oo, teq_terminate))
		{
			/* mark this octets as being shutdown/terminated */
			/* when the opposite makes it's way through here, it will be
			 * point_xrelease. */
			Point_SetKPoint(o, -1);
			Transit_XQualify(o, teq_terminate);
			return;
		}
	}

	return(ti_point_xrelease(t));
}

static void
ti_octets_irelease(Transit t)
{
	Octets o = (Octets) t;

	if (o->resource != NULL)
	{
		o->last_position = o->position = -1;
		PyBuffer_Release(&(o->view));
		Py_DECREF(o->resource);
		o->resource = NULL;
	}
}

struct TransitInterface OctetsTIF = {
	ti_point_kfilter,
	ti_point_transform,
	ti_octets_transfer,
	ti_octets_terminate,
	ti_octets_xrelease,
	ti_octets_irelease,
};

/*
 * deallocation is straight forward except in the case of sockets,
 * which have a shared file descriptor and must refer to each other
 * to identify whether or not the kpoint can be closed.
 */
static void
octets_dealloc(PyObj self)
{
	Octets o = (Octets) self;

	if (o->position != -1)
	{
		Py_DECREF(o->resource);
		o->resource = NULL;
	}
	o->position = o->last_position = -1;

	/*
	 * For sockets, the receiving Octets instance
	 * must be deallocated first; it will hold an additional
	 * reference to sending Octets.
	 */
	if (Octets_IsSocket(o) && Transit_Receives(o))
	{
		Octets oo = Octets_GetOpposite(o);
		kpoint_t kp;

		/* signals the write side that the receive side is gone */
		Octets_SetOpposite(oo, NULL);
		kp = Point_GetKPoint(oo);
		Py_DECREF(oo);

		if (kp != -1)
		{
			/*
			 * Opposite's KP is valid, let the sends transit xrelease the KP.
			 */
			Point_SetKPoint(o, -1);
			return(transit_dealloc(self));
		}
		/*
		 * Otherwise, the opposite has been deallocated/xrelease'd, leaving this
		 * receiver to terminate xrelease the shared kpoint. However, this receiver
		 * is being dealloc'd and the opposite (oo) already performed
		 */
	}

	return(point_dealloc(self));
}

const char octets_doc[] =
"Octets is a Transit that transfers 'octets', binary data. Primarily, data\n"
"transfers are facilitated by buffer-capable objects such a :py:class:`bytearrays`\n"
"or :py:class:`bytes` instances."
;
TransitPyTypeObject OctetsType = {{
	PyVarObject_HEAD_INIT(NULL, 0)
	QPATH("Octets"),					/* tp_name */
	sizeof(struct Octets),			/* tp_basicsize */
	0,										/* tp_itemsize */
	octets_dealloc,					/* tp_dealloc */
	NULL,									/* tp_print */
	NULL,									/* tp_getattr */
	NULL,									/* tp_setattr */
	NULL,									/* tp_compare */
	NULL,									/* tp_repr */
	NULL,									/* tp_as_number */
	NULL,									/* tp_as_sequence */
	NULL,									/* tp_as_mapping */
	NULL,									/* tp_hash */
	NULL,									/* tp_call */
	NULL,									/* tp_str */
	NULL,									/* tp_getattro */
	NULL,									/* tp_setattro */
	NULL,									/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT,	  			/* tp_flags */
	octets_doc,							/* tp_doc */
	NULL,									/* tp_traverse */
	NULL,									/* tp_clear */
	NULL,									/* tp_richcompare */
	0,										/* tp_weaklistoffset */
	NULL,									/* tp_iter */
	NULL,									/* tp_iternext */
	octets_methods,					/* tp_methods */
	octets_members,					/* tp_members */
	octets_getset,						/* tp_getset */
	&PointType.typ,					/* tp_base */
	NULL,									/* tp_dict */
	NULL,									/* tp_descr_get */
	NULL,									/* tp_descr_set */
	0,										/* tp_dictoffset */
	NULL,									/* tp_init */
	NULL,									/* tp_alloc */
	NULL,									/* tp_new */
},
	&OctetsTIF
};

static void
ti_sockets_transfer(Transit t)
{
	Sockets S = (Sockets) t;
	struct sockaddr addrinfo;
	socklen_t addrlen;
	kpoint_t kp = -1;
	kerror_t ke = 0;
	int i = 0;
	int *accepted = Sockets_GetArray(S);
	Py_ssize_t remainder = Sockets_GetRemainder(S);
	Py_ssize_t max = MIN(remainder, S->ksize);

	/*
	 * Accept loop. Read as many as we can.
	 */
	while (i < max)
	{
		kp = accept(Point_GetKPoint(S), &addrinfo, &addrlen);

		if (kp < 0)
		{
			ke = errno;

			switch (ke)
			{
				/*
				 * All of these cases do not indicate fatality
				 */
				case EMFILE:
				case ENFILE:
				case ENOMEM:
				case EWOULDBLOCK:
					/*
					 * Note memory problems.
					 */
					Point_SetKCall(S, kc_accept);
					Point_SetKError(S, ke);
				case EINTR: /* However, don't bother reporting EINTR. */
					/*
					 * This can happen in cases where multiple processes
					 * are accepting connections on the same socket.
					 */
					goto exit;
				break;

				default:
					Point_SetKCall(S, kc_accept);
					Point_SetKError(S, ke);
					Transit_XQualify(S, teq_terminate);
				break;
			}
		}
		else
		{
			accepted[i] = (int) kp;
			++i;
		}
	}

exit:
	S->last_position = S->position;
	S->position = S->last_position + i;
	if (i > 0)
		Transit_NoteEvent(S, tev_transfer);

	if (i == remainder)
	{
		Transit_INQualify(S, teq_transfer);
		Transit_NoteEvent(S, tev_exhaust);
	}

	if (i == S->ksize)
	{
		/*
		 * Exhausted the last report of available connections.
		 */
		Transit_XNQualify(S, teq_transfer);
	}
}

static void
ti_sockets_irelease(Transit t)
{
	Sockets S = (Sockets) t;

	if (S->resource != NULL)
	{
		S->last_position = S->position = -1;
		PyBuffer_Release(&(S->view));
		Py_DECREF(S->resource);
		S->resource = NULL;
	}
}

struct TransitInterface SocketsTIF = {
	ti_point_kfilter,
	ti_point_transform,
	ti_sockets_transfer,
	ti_transit_terminate,
	ti_point_xrelease,
	ti_sockets_irelease,
};

static PyObj
sockets_identity(PyObj self)
{
	return(socket_identity((Point) self, 0));
}

static PyObj
sockets_transfer(PyObj self)
{
	Sockets S = (Sockets) self;
	return(_PySlice_FromIndices(S->last_position, S->position));
}

static PyObj
sockets_acquire(PyObj self, PyObj resource)
{
	Sockets S = (Sockets) self;

	if (!transit_can_acquire((Transit) S))
		return(NULL);

	if (S->position >= 0)
	{
		/*
		 * Transit_IQualified(o, teq_transfer) should be FALSE.
		 *
		 * Release the buffer and reset the position information so in
		 * error cases below, there isn't a chance that Release will be ran
		 * twice upon re-entry.
		 */
		ti_sockets_irelease((Transit) S);
	}


	S->resource = resource;
	Py_INCREF(S->resource);

	if (PyObject_GetBuffer(resource, &(S->view), PyBUF_WRITABLE))
	{
		Py_DECREF(S->resource);
		S->resource = NULL;
		return(NULL);
	}

	S->last_position = S->position = 0;

	transit_acquired((Transit) S);
	Py_INCREF(self);
	return(self);
}

static void
scbl(int family, int socktype,
	struct sockaddr *bind_address, size_t basize,
	int backlog,
	kpoint_t *kp, kcall_t *kc, kerror_t *ke)
{
	*kp = socket(family, socktype, 0);

	if (*kp < 0)
		*kc = kc_socket;
	else
	{
		init_socket((*kp)); /* configures no blocking, sigpipe, etc */

		if (bind((int) (*kp), bind_address, basize))
			*kc = kc_bind;
		else
		{
			if (listen((int) *kp, backlog))
				*kc = kc_listen;
			else
				return; /* everything succeeded */
		}
	}

	*ke = errno;
}

/*
 * Listen for IPv4 connections.
 */
static PyObj
accept_stream_ip4(PyObj subtype,
	PyObj bind_addr_ob, unsigned short bind_port,
	int socktype, int backlog)
{
	Sockets S;
	struct sockaddr_in bind_addr = {0,};

	kpoint_t kp = -1;
	kcall_t kc = kc_none;
	kerror_t ke = 0;

	switch (inet_from_pystr(&(bind_addr.sin_addr.s_addr), PF_INET, bind_addr_ob))
	{
		case 0:
			/* got it */
		break;

		case -1:
			/* error */
			return(NULL);
		break;

		case 1:
		{
			/* not processed / not a string or buffer */
			if (inet4_from_pyint(&(bind_addr.sin_addr.s_addr), bind_addr_ob))
				return(NULL);

			if (PyErr_Occurred())
				return(NULL);
		}
		break;
	}

	S = (Sockets) PyAllocate(subtype);
	if (S == NULL)
		return(NULL);

	Py_BEGIN_ALLOW_THREADS
	{
		bind_addr.sin_port = htons(bind_port);
		bind_addr.sin_family = PF_INET;

		scbl(PF_INET, socktype,
			(struct sockaddr *) &bind_addr, sizeof(bind_addr),
			backlog, &kp, &kc, &ke);
		INIT_SOCKETS(S, kp, kc, ke);
	}
	Py_END_ALLOW_THREADS

	return((PyObj) S);
}

/*
 * Listen for IPv6 connections.
 */
static PyObj
accept_stream_ip6(PyObj subtype,
	PyObj bind_addr_ob, unsigned short bind_port,
	int socktype, int backlog)
{
	Sockets S;
	struct sockaddr_in6 bind_addr = {0,};

	kpoint_t kp = -1;
	kcall_t kc = kc_none;
	kerror_t ke = 0;

	switch (inet_from_pystr(&(bind_addr.sin6_addr), PF_INET6, bind_addr_ob))
	{
		case 0:
			/* got it */
		break;

		case -1:
			/* error */
			return(NULL);
		break;

		case 1:
		{
			/* not processed / not a string or buffer */
			if (inet6_from_pyint(&(bind_addr.sin6_addr), bind_addr_ob))
				return(NULL);

			if (PyErr_Occurred())
				return(NULL);
		}
		break;
	}

	S = (Sockets) PyAllocate(subtype);
	if (S == NULL)
		return(NULL);

	Py_BEGIN_ALLOW_THREADS
	{
		bind_addr.sin6_family = PF_INET6;
		bind_addr.sin6_port = htons(bind_port);

		scbl(PF_INET6, socktype,
			(struct sockaddr *) &bind_addr, sizeof(bind_addr), backlog, &kp, &kc, &ke);
		INIT_SOCKETS(S, kp, kc, ke);
	}
	Py_END_ALLOW_THREADS

	return((PyObj) S);
}

static PyObj
accept_stream_local(PyObj subtype, char *path, int socktype, int backlog)
{
	Sockets S;
	struct sockaddr_un bind_addr;

	kpoint_t kp = -1;
	kcall_t kc = kc_none;
	kerror_t ke = 0;

	S = (Sockets) PyAllocate(subtype);
	if (S == NULL)
		return(NULL);

	Py_BEGIN_ALLOW_THREADS
	{
		strncpy((char *) &(bind_addr.sun_path), path, sizeof(bind_addr.sun_path));
		bind_addr.sun_family = PF_LOCAL;

		scbl(PF_LOCAL, socktype,
			(struct sockaddr *) &bind_addr, sizeof(bind_addr),
			backlog, &kp, &kc, &ke);
		INIT_SOCKETS(S, kp, kc, ke);
	}
	Py_END_ALLOW_THREADS

	return((PyObj) S);
}

static PyObj
accept_stream_socket(PyObj subtype, kpoint_t kp, int backlog)
{
	Sockets S;
	kcall_t kc = kc_none;
	kerror_t ke = 0;

	S = (Sockets) PyAllocate(subtype);
	if (S == NULL)
		return(NULL);

	Py_BEGIN_ALLOW_THREADS
	{
		init_socket(kp);

		if (backlog)
		{
			if (!listen(kp, backlog))
			{
				ke = errno;
				kc = kc_listen;
			}
		}

		INIT_SOCKETS(S, kp, kc, ke);
	}
	Py_END_ALLOW_THREADS

	return((PyObj) S);
}

/*
 * traffic.libkqueue.Sockets.accept_stream_ip4(if4, port, backlog = DEFAULT_BACKLOG)
 */
static PyObj
sockets_accept_stream_ip4(PyObj subtype, PyObj args, PyObj kw)
{
	static char *kwlist[] = {"interface", "backlog", NULL};
	PyObj bind_addr_ob = NULL;
	unsigned short bind_port = 0;
	int backlog = DEFAULT_BACKLOG;

	if (!PyArg_ParseTupleAndKeywords(args, kw, "(OH)|i", kwlist, &bind_addr_ob, &bind_port, &backlog))
		return(NULL);

	return(accept_stream_ip4(subtype, bind_addr_ob, bind_port, SOCK_STREAM, backlog));
}

/*
 * traffic.libkqueue.Sockets.accept_stream_ip6((if6, port), backlog = 128)
 */
static PyObj
sockets_accept_stream_ip6(PyObj subtype, PyObj args, PyObj kw)
{
	static char *kwlist[] = {"interface", "backlog", NULL};
	PyObj bind_addr_ob = NULL;
	unsigned short bind_port = 0;
	int backlog = DEFAULT_BACKLOG;

	if (!PyArg_ParseTupleAndKeywords(args, kw, "(OH)|i", kwlist, &bind_addr_ob, &bind_port, &backlog))
		return(NULL);

	return(accept_stream_ip6(subtype, bind_addr_ob, bind_port, SOCK_STREAM, backlog));
}

/*
 * traffic.libkqueue.Sockets.accept_stream_local(fspath)
 */
static PyObj
sockets_accept_stream_local(PyObj subtype, PyObj args, PyObj kw)
{
	static char *kwlist[] = {"path", "backlog", NULL};
	char *path;
	int backlog = DEFAULT_BACKLOG;

	if (!PyArg_ParseTupleAndKeywords(args, kw, "y|i", kwlist, &path, &backlog))
		return(NULL);

	return(accept_stream_local(subtype, path, SOCK_STREAM, backlog));
}

/*
 * traffic.libkqueue.Sockets.accept_stream_socket()
 */
static PyObj
sockets_accept_stream_socket(PyObj subtype, PyObj args, PyObj kw)
{
	static char *kwlist[] = {"fileno", "backlog", NULL};
	int fileno, backlog = DEFAULT_BACKLOG;

	if (!PyArg_ParseTupleAndKeywords(args, kw, "i|i", kwlist, &fileno, &backlog))
		return(NULL);

	return(accept_stream_socket(subtype, (kpoint_t) fileno, backlog));
}

static PyObj
sockets_rallocate(PyObj subtype, PyObj args)
{
	return(PyObject_Call(new_array, args, NULL));
}

/* METH_O, METH_VARARGS, METH_VARKEYWORDS, METH_NOARGS */
static PyMethodDef sockets_methods[] = {
	{"rallocate",
		(PyCFunction) sockets_rallocate,
		METH_VARARGS|METH_CLASS,
		PyDoc_STR(
"rallocate(size)\n\n"
":param size: Number of sockets that can be accepted.\n"
":type size: :py:class:`int`\n"
":returns: A new mutable resource.\n"
":rtype: :py:class:`array.array`\n"
"\n"
"Create a mutable resource capable of being written into by a Sockets instance.\n"
)},

/* creation */
	{"accept_stream_ip4",
		(PyCFunction) sockets_accept_stream_ip4,
		METH_VARARGS|METH_KEYWORDS|METH_CLASS,
		PyDoc_STR(
"accept_stream_ip4(interface, backlog = " DEFAULT_BACKLOG_STR ")\n\n"
":param interface: The IPv4 interface and port pair to bind to.\n"
":type interface: ([:py:class:`str` or :py:class:`int`], :py:class:`int`)\n"
":param backlog: The listen backlog to configure.\n"
":type backlog: int\n"
":returns: A new Sockets instance.\n"
":rtype: " DRC(Sockets) "\n"
"\n"
"Create a Sockets instance that accepts sockets for IPv4 (PF_INET) stream connections.\n"
)},

	{"accept_stream_ip6",
		(PyCFunction) sockets_accept_stream_ip6,
		METH_VARARGS|METH_KEYWORDS|METH_CLASS,
		PyDoc_STR(
"accept_stream_ip6(interface, backlog = " DEFAULT_BACKLOG_STR ")\n\n"
":param interface: The IPv6 interface and port pair to bind to.\n"
":type interface: ([:py:class:`str` or :py:class:`int`], :py:class:`int`)\n"
":param backlog: The listen backlog to configure.\n"
":type backlog: :py:class:`int`\n"
":returns: A new Sockets instance.\n"
":rtype: " DRC(Sockets) "\n"
"\n"
"Create a Sockets instance that accepts sockets for IPv6 (PF_INET6) stream connections.\n"
)},

	/* identifies it as AF_UNIX, signalling unlink on terminate */
	{"accept_stream_local",
		(PyCFunction) sockets_accept_stream_local,
		METH_VARARGS|METH_KEYWORDS|METH_CLASS,
		PyDoc_STR(
"accept_stream_local(interface, backlog = " DEFAULT_BACKLOG_STR ")\n\n"
":param interface: The file system path to bind to. For example, ``b'/tmp/server'``.\n"
":type interface: :py:class:`bytes`\n"
":param backlog: The listen backlog to configure.\n"
":type backlog: :py:class:`int`\n"
":return: A new Sockets instance.\n"
":rtype: " DRC(Sockets) "\n"
"\n"
"Create a Sockets instance that accepts sockets for local (PF_UNIX) stream connections.\n"
"\n"
".. note:: The file system path *must* be in bytes. It is necessary to properly encode "
"the parameter into the file system's encoding.\n"
)},

	{"accept_stream_socket",
		(PyCFunction) sockets_accept_stream_socket,
		METH_VARARGS|METH_KEYWORDS|METH_CLASS,
		PyDoc_STR(
"accept_stream_socket(fileno, backlog = " DEFAULT_BACKLOG_STR ")\n\n"
":param fileno: The socket file descriptor to accept connections on.\n"
":type fileno: :py:class:`int`\n"
":param backlog: The listen backlog to configure.\n"
":type backlog: :py:class:`int`\n"
":returns: A new Sockets instance.\n"
":rtype: " DRC(Sockets) "\n"
"\n"
"Create a Sockets instance for accepting new sockets from the given file descriptor.\n"
":manpage:`listen(2)` will be called unless a zero-backlog is given.\n"
"*The file descriptor will have its options modified*."
)},

	{"identity",
		(PyCFunction) sockets_identity,
		METH_NOARGS,
		PyDoc_STR(
"identity()\n\n"
":returns: Interface-port pair that the socket is bound to.\n"
":rtype: (:py:class:`str`, :py:class:`int`)\n"
"\n"
"Builds a tuple identifying the interface that is accepting connections."
)},

	{"transfer",
		(PyCFunction) sockets_transfer,
		METH_NOARGS,
		PyDoc_STR(
"transfer()\n\n"
":returns: Return a pair of integers specifying the index range that the accepted file descriptors were written to.\n"
":rtype: :py:class:`slice`\n"
"\n"
)},

	{"acquire",
		(PyCFunction) sockets_acquire,
		METH_O,
		PyDoc_STR(
"acquire(mutable_buffer)\n\n"
":param mutable_buffer: The area of memory to write the socket identifiers to.\n"
":type mutable_buffer: buffer\n"
":returns: The Sockets instance acquiring the resource. (Method Chaining)\n"
":rtype: " DRC(Sockets) "\n"
"\n"
"Acquire a resource for storing accepted file descriptors. Usually, " DRM(Octets.rallocate) " is used to allocate these resources.\n"
)},
	{NULL,}
};

static PyMemberDef sockets_members[] = {
	{"resource", T_OBJECT, offsetof(struct Sockets, resource), READONLY,
		PyDoc_STR(
"The object whose buffer was acquired, " DRM(Sockets.acquire) ", as the Transit's transfer resource.\n\n:py:obj:`None` if there is no resource."
)},
	{NULL,},
};

static PyGetSetDef sockets_getset[] = {
	{NULL,},
};

const char sockets_doc[] =
"Sockets is a Point based Transit transferring accepted socket "
"file descriptors from a listening socket."
;
TransitPyTypeObject SocketsType = {{
	PyVarObject_HEAD_INIT(NULL, 0)
	QPATH("Sockets"),				/* tp_name */
	sizeof(struct Sockets),		/* tp_basicsize */
	0,									/* tp_itemsize */
	NULL,								/* tp_dealloc */
	NULL,								/* tp_print */
	NULL,								/* tp_getattr */
	NULL,								/* tp_setattr */
	NULL,								/* tp_compare */
	NULL,								/* tp_repr */
	NULL,								/* tp_as_number */
	NULL,								/* tp_as_sequence */
	NULL,								/* tp_as_mapping */
	NULL,								/* tp_hash */
	NULL,								/* tp_call */
	NULL,								/* tp_str */
	NULL,								/* tp_getattro */
	NULL,								/* tp_setattro */
	NULL,								/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT,	  		/* tp_flags */
	sockets_doc,					/* tp_doc */
	NULL,								/* tp_traverse */
	NULL,								/* tp_clear */
	NULL,								/* tp_richcompare */
	0,									/* tp_weaklistoffset */
	NULL,								/* tp_iter */
	NULL,								/* tp_iternext */
	sockets_methods,				/* tp_methods */
	sockets_members,				/* tp_members */
	sockets_getset,				/* tp_getset */
	&PointType.typ,				/* tp_base */
	NULL,								/* tp_dict */
	NULL,								/* tp_descr_get */
	NULL,								/* tp_descr_set */
	0,									/* tp_dictoffset */
	NULL,								/* tp_init */
	NULL,								/* tp_alloc */
	NULL,								/* tp_new */
},
	&SocketsTIF,
};

static PyObj
traffic_acquire(PyObj self, PyObj ob)
{
	Traffic T = (Traffic) self;
	Transit t = (Transit) ob;

	if (!PyObject_IsInstance(ob, (PyObj) &TransitType))
	{
		PyErr_SetString(PyExc_TypeError, "cannot attach objects that are not transits");
		return(NULL);
	}

	if (Transit_Terminating(T))
	{
		/* Traffic is Terminated */
		PyErr_SetTransitTerminatedError((Transit) T);
		return(NULL);
	}

	if (Transit_GetTraffic(t) == NULL)
	{
		if (Transit_Terminated(t))
		{
			/* Given Transit is Terminated */
			/*
			 * Terminating check performed after the NULL check because
			 * if the given transit is already acquired by the Traffic,
			 * it shouldn't complain due to it already being acquired.
			 *
			 * Additionally, it's desired that ResourceError is consistently
			 * thrown in this case.
			 */
			PyErr_SetTransitTerminatedError((Transit) t);
			return(NULL);
		}

		/* Special event that identifies the new Transit. */
		Transit_NoteEvent(t, tev_traffic);

		Py_INCREF(T); /* Newly acquired transit's reference to Traffic. */
		Py_INCREF(t); /* Traffic's reference to the newly acquired Transit. */

		Transit_SetTraffic(t, T);
		TRANSIT_ATTACH(t);

		Traffic_IncrementTransitCount(T);
	}
	else
	{
		if (Transit_GetTraffic(t) != T)
		{
			/*
			 * Another Traffic instance acquired the Transit.
			 */
			PyErr_SetString(PyExc_ResourceError, "transit attached to distinct traffic");
			return(NULL);
		}

		/*
		 * Otherwise, just fall through as it's already acquired.
		 */
	}

	Py_INCREF(ob);
	return(ob);
}

static kerror_t
traffic_init_kqueue(Traffic T)
{
	int fd;
	struct kevent kev;
	struct timespec ts = {0,};

	fd = kqueue();
	Point_SetKPoint(T, (kpoint_t) fd);
	if (fd < 0)
	{
		Point_SetKError(T, (kerror_t) errno);
		Point_SetKCall(T, kc_kqueue);
		return(errno);
	}

	kev.udata = (void *) T;
	kev.ident = (uintptr_t) T;
	kev.flags = EV_ADD|EV_RECEIPT|EV_CLEAR;
	kev.filter = EVFILT_USER;
	kev.fflags = 0;
	kev.data = 0;

	/*
	 * Setup the user trigger for Traffic.force().
	 */
	if (kevent(Point_GetKPoint(T), &kev, 1, &kev, 1, &ts) < 0)
	{
		kerror_t err = errno;
		Point_SetKPoint(T, -1);
		Point_SetKError(T, err);
		Point_SetKCall(T, kc_kevent);
		close(fd);
		return(err);
	}

	return(0);
}

static void
traffic_start_cycle(Traffic T)
{
	T->lltransfer = (Transit) T; /* Marks that it's in a cycle. */
}

static void
traffic_finish_cycle(Traffic T, unsigned long count)
{
	T->lltransfer = (Transit) NULL; /* Marks that it's done with the cycle. */
	T->lltransfer_ptr = (Transit) T;
	T->transfer_count = 0;

	/*
	 * Complete Traffic termination?
	 */
	if (Transit_Terminating(T))
		Transit_XQualify(T, teq_terminate);
}

static void
traffic_kevent_change(Traffic T)
{
	int r;
	pkevents("changing", T);

dokevent:
	r = kevent(Point_GetKPoint(T), T->kevs, (int) T->ksize, T->kevs, (int) T->ksize, NULL);
	if (r == -1)
	{
		r = errno;

		switch (r)
		{
			case EINTR:
				/*
				 * Try again and again.
				 * Changes *must* happen as cycles don't keep track of what has
				 * been moved on to the changes array.
				 */
				goto dokevent;
			break;

			case EBADF:
				/*
				 * Set to -1 to signal re-initialization.
				 */
				Point_SetKPoint(T, -1);
			break;

			case ENOENT:
			case ENOMEM:
			default:
				Point_SetKError(T, r);
				Point_SetKCall(T, kc_kevent);
			break;
		}

	}
	T->ksize = 0;
}

static void
traffic_kevent_collect(Traffic T, char poll)
{
	const static struct timespec nowait = {0,0};
	const static struct timespec waitfor = {2,0};
	struct timespec *wait = (struct timespec *) (poll ? &nowait : &waitfor);

	T->ksize = kevent(Point_GetKPoint(T), NULL, 0, T->kevs, N_KEVENTS, wait);
	if (T->ksize == -1)
	{
		switch (errno)
		{
			case EBADF:
				Point_SetKPoint(T, -1);
			case EINTR:
				/*
				 * Do *not* try again.
				 * This is an interrupt on collection.
				 * The cycle will safely exit as if no events were picked up, and
				 * there is a good chance that an interrupt needs to be serviced.
				 */
				T->ksize = 0;
			break;

			case ENOENT:
			case ENOMEM:
			default:
				Point_SetKError(T, (kerror_t) errno);
				Point_SetKCall(T, kc_kevent);
			break;
		}
	}
	pkevents("collected", T);
}

/*
 * Note tev_traffic events on all Transits.
 *
 * Run before traffic_transfer_delta to have all
 * Transit's corresponding kevent filter to be loaded.
 */
static void
traffic_reload(Traffic T)
{
	Transit t = T->next;

	/* MUST HAVE GIL */

	while (t != (Transit) T)
	{
		Transit_NoteEvent(t, tev_traffic);
		t = t->next;
	}
}

static void
traffic_transfer_delta(Traffic T)
{
	Transit t;

	/* MUST HAVE GIL */
	for (t = T->prev; Transit_GetDelta(t) != 0 || Transit_ShouldXConnect(t); t = t->prev)
	{
		/*
		 * prepend to the lltransfer list.
		 * The first 't' was the last enqueued.
		 */
		Transit_AddEquals(t, Transit_GetDelta(t)); /* Record the internal event quals. */
		Transit_ClearDelta(t); /* Clear delta for subsequent use. */
		t->lltransfer = t->next; /* Place the transit on the transfer linked list. */
		++(T->transfer_count);
	}

	/*
	 * Did anything get prepended?
	 */
	if (T->prev->lltransfer != NULL)
	{
		T->lltransfer = t->next;
		T->lltransfer_ptr = T->prev;
	}
}

/*
 * transit_xdisconnect - stop listening for events from the kernel
 */
static void
transit_xdisconnect(Transit t, kevent_t kev)
{
	TransitInterface tif = Transit_GetInterface(t);

	if (tif->ti_kfilter != NULL)
	{
		tif->ti_kfilter(t, kev);
		kev->flags |= EV_DELETE|EV_RECEIPT;
		kev->udata = t;
	}
}

/*
 * transit_xconnect - listen for events from the kernel
 */
static void
transit_xconnect(Transit t, kevent_t kev)
{
	TransitInterface tif = Transit_GetInterface(t);

	if (tif->ti_kfilter != NULL)
	{
		tif->ti_kfilter(t, kev);
		kev->flags |= EV_ADD|EV_RECEIPT;
		kev->udata = t;
	}
}

static kevent_t
traffic_next_kevent_slot(Traffic T)
{
	kevent_t kev;

	if (T->ksize == N_KEVENTS)
	{
		traffic_kevent_change(T);
		T->ksize = 0;
	}

	kev = &(T->kevs[T->ksize]);
	++T->ksize;

	return(kev);
}

/*
 * Process signals and setup for event processing if possible.
 */
static void
traffic_apply_delta(Traffic T)
{
	Transit prev, t;
	kevent_t kev;

	/* NO GIL */

	prev = (Transit) T;
	for (t = T->lltransfer; t != (Transit) T; t = prev->lltransfer)
	{
		if (Transit_ShouldXConnect(t))
		{
			/*
			 * iff xterminate hasn't occurred.
			 * (happens with transits that are terminated at creation due to fail)
			 */
			if (!Transit_XQualified(t, teq_terminate))
			{
				kev = traffic_next_kevent_slot(T);
				transit_xconnect(t, kev);
			}

			Transit_ClearEvents(t); /* denoted "ShouldXConnect" */
		}

		if (Transit_ShouldTransfer(t) || Transit_IQualified(t, teq_terminate))
		{
			/*
			 * There is no check for "should exhaust" as exhaustion only
			 * follows a transfer. Exhaust events are determined after the transfer
			 * is emitted.
			 */
			prev = t;
		}
		else
		{
			/*
			 * Incomplete qualifications, remove transit from list.
			 */
			prev->lltransfer = t->lltransfer;
			t->lltransfer = NULL;
			/*
			 * It was just a signal; decrement the transfer_count.
			 */
			--(T->transfer_count);
		}
	}
	T->lltransfer_ptr = prev;
}

/*
 * Transform the collected events into local Transit state.
 */
static void
traffic_kevent_transform(Traffic T, kevent_t kevs)
{
	Transit t;
	kevent_t kev;
	int i;

	/*
	 * Iterate over the kevent's collected events and
	 * transform the kevent state data into Transit state.
	 */
	for (i = 0; i < T->ksize; ++i)
	{
		kev = &(T->kevs[i]);
		t = (Transit) kev->udata;

		/*
		 * (EVFILT_USER) user signaled for kevent exit?
		 */
		if (t == (Transit) T)
			continue;

		Transit_GetInterface(t)->ti_transform(t, kev);
		if (Transit_ShouldTransfer(t) || Transit_ShouldTerminate(t))
			Transit_EnqueueEvent(t);
	}
}

/*
 * traffic.libkqueue.Traffic.force()
 */
static PyObj
traffic_force(PyObj self)
{
	Traffic T = (Traffic) self;
	struct timespec ts = {0,0};
	struct kevent kev;

	kev.udata = (void *) self;
	kev.ident = (uintptr_t) self;
	kev.filter = EVFILT_USER;
	kev.fflags = NOTE_TRIGGER;
	kev.data = 0;
	kev.flags = EV_RECEIPT;

	if (kevent(Point_GetKPoint(T), &kev, 1, NULL, 0, &ts) < 0)
	{
		PyErr_SetFromErrno(PyExc_OSError);
		return(NULL);
	}

	Py_INCREF(self);
	return(self);
}

static void
ti_traffic_xrelease(Transit T)
{
	return(ti_point_xrelease(T));
}

static void
ti_traffic_terminate(Transit T)
{
	Transit t;

	Transit_IQualify(T, teq_terminate);

	ptransit(T);
	for (t = T->next; t != T; t = t->next)
	{
		/*
		 * Enqueue is necessary here because ALL transits will
		 * have a terminate action.
		 */
		Transit_DQualify(t, teq_terminate);
	}

	ti_transit_terminate(T);
	return(ti_traffic_xrelease(T));
}

/*
 * _traffic_flow() - collect and process traffic events
 */
static void
_traffic_flow(Traffic T)
{
	Transit t;

	traffic_start_cycle(T);

	/*
	 * Check for Traffic termination.
	 */
	if (Transit_Terminating(T))
	{
		/*
		 * terminate all transits
		 */
		ti_traffic_terminate((Transit) T);
	}
	else if (Point_GetKPoint(T) == -1)
	{
		/*
		 * kqueue file descriptor went bad.
		 * Either a fork occurred or the user close()'d it.
		 */
		traffic_init_kqueue(T);
		traffic_reload(T);
	}
	Transit_ClearDelta(T);

	/*
	 * Enqueue changed transits to lltransfer while holding the GIL.
	 */
	traffic_transfer_delta(T);

	/* 
	 * The GIL is no longer necessary, and concurrent
	 * code can send signals Transits as desired.
	 */

	Py_BEGIN_ALLOW_THREADS

	/*
	 * t->next/t->prev CAN BE USED BY OTHER THREADS. DO NOT USE WITHOUT GIL.
	 */

	traffic_apply_delta(T);
	traffic_kevent_change(T); /* make any remaining connections */

	/* don't bother collecting/transforming if terminating */
	if (!Transit_Terminating(T))
	{
		if (T->lltransfer == (Transit) T)
		{
			/*
			 * Collect and transform given that there are no signal-queued transits
			 * or that the number of cycles is evenly divisible by the collection
			 * frequency, limiting the number of times that a cycle can occur
			 * without collecting new events.
			 */
			traffic_kevent_collect(T, 0);
			traffic_kevent_transform(T, T->kevs);

			/*
			 * Iff more kevents may exists, collect more until the event count exceeds 75%
			 */
			while (T->ksize == N_KEVENTS && T->transfer_count < ((T->transit_count * 3) / 4))
			{
				traffic_kevent_collect(T, 1);
				traffic_kevent_transform(T, T->kevs);
			}
			T->ksize = 0;
		}
	}

	/*
	 * Iterate over all the transits where an event is
	 * a possibility.
	 */
	for (t = T->lltransfer; t != (Transit) T; t = t->lltransfer)
	{
		TransitInterface tif = Transit_GetInterface(t);

		ptransit(t);

		if (Transit_ShouldTerminate(t))
		{
			/*
			 * ti_terminate may scribble on the transit,
			 * so disconnect first.
			 */
			transit_xdisconnect(t, traffic_next_kevent_slot(T));
			if (tif->ti_terminate)
				tif->ti_terminate(t);
			Transit_NoteEvent(t, tev_terminate);
		}
		else if (Transit_ShouldTransfer(t))
		{
			tif->ti_transfer(t);
		}
	}

	/*
	 * Perform any disconnects queued up in the loop.
	 */
	if (!Transit_Terminating(T) && T->ksize > 0)
		traffic_kevent_change(T);

	Py_END_ALLOW_THREADS
}

static void
ti_traffic_transform(Transit t, kevent_t kev)
{
	return;
}

static void
ti_traffic_transfer(Transit t)
{
	return;
}

struct TransitInterface TrafficTIF = {
	NULL,
	ti_traffic_transform,
	ti_traffic_transfer,
	ti_traffic_terminate,
	ti_transit_xrelease,
};

/*
 * _traffic_events - return an iterable to the collected events.
 */
static PyObj
_traffic_events(Traffic T)
{
	if (!Transit_InCycle(T))
	{
		PyErr_SetString(PyExc_RuntimeError, "no events outside of a cycle");
	}

	return(new_lltransfer_iter(T));
}

/*
 * Traffic.transfer() - return an iterable to the collected events.
 */
static PyObj
traffic_transfer(PyObj self)
{
	return(_traffic_events((Traffic) self));
}

/*
 * Traffic.disregard()
 */
static PyObj
traffic_disregard(PyObj self)
{
	Transit t, T = (Transit) self;

	/* GIL Required */

	for (t = T->next; t != T; t = t->next)
	{
		PyObj tob = (PyObj) transit_disregard((PyObj) t);
		Py_DECREF(tob);
	}

	return(transit_terminate(self));
}

static unsigned long
_traffic_flush(Traffic T)
{
	unsigned long count = 0;
	Transit t, next;

	/* REQUIRES GIL */
	t = T->lltransfer;
	while (t != (Transit) T)
	{
		++count;
		next = t->lltransfer;
		t->lltransfer = NULL;

		if (Transit_HasEvent(t, tev_exhaust) && !Transit_DQualified(t, tev_transfer))
		{
			/*
			 * Exhaust event occurred, but no new resource supplied in cycle.
			 * Release any internal resources.
			 */
			Transit_GetInterface(t)->ti_irelease(t);
		}

		if (Transit_HasEvent(t, tev_terminate))
		{
			/*
			 * Release any resources owned by the transit.
			 */
			Transit_GetInterface(t)->ti_irelease(t);
			Transit_GetInterface(t)->ti_xrelease(t);

			/*
			 * Clear the link, helping break any (gc) cycles.
			 */
			Py_XDECREF(t->link);
			t->link = NULL;

			TRANSIT_DETACH(t);
			Traffic_DecrementTransitCount(Transit_GetTraffic(t));

			/*
			 * Emitted termination? Release traffic's reference to the transit.
			 */
			Py_DECREF(t);
			ptransit(t);
		}

		Transit_ClearEvents(t);
		t = next;
	}

	traffic_finish_cycle(T, count);
	return(count);
}

/*
 * Traffic.__enter__() - enter the cycle
 */
static PyObj
traffic_enter(PyObj self)
{
	Traffic T = (Traffic) self;

	if (Transit_Terminated(T))
	{
		PyErr_SetTransitTerminatedError(T);
		return(NULL);
	}

	if (Transit_InCycle(T))
	{
		PyErr_SetString(PyExc_RuntimeError,
			"cycle must be completed before starting another");
		return(NULL);
	}

	_traffic_flow(T);

	Py_INCREF(self);
	return(self);
}

/*
 * Traffic.__exit__() - exit the cycle
 */
static PyObj
traffic_exit(PyObj self, PyObj args)
{
	Traffic T = (Traffic) self;

	if (!Transit_InCycle(T))
	{
		PyErr_SetString(PyExc_RuntimeError, "cannot flush outside of cycle");
		return(NULL);
	}

	_traffic_flush(T);
	Py_RETURN_NONE;
}

/* METH_O, METH_VARARGS, METH_VARKEYWORDS, METH_NOARGS */
static PyMethodDef traffic_methods[] = {
	{"acquire", (PyCFunction) traffic_acquire,
		METH_O, PyDoc_STR(
"acquire(transit)\n\n"
":param transit: The Transit instance to place In Traffic.\n"
":type transit: " DRC(Transit) "\n"
"\n"
"Acquires the Transit in order to enable the Transits' events to be collected in\n"
"the "DRC(Traffic)"'s cycles."
)},

	{"disregard", (PyCFunction) traffic_disregard,
		METH_NOARGS, PyDoc_STR(
"disregard()\n\n"
":returns: The Traffic instance being disregarded. (Method Chaining)\n"
":rtype: " DRC(Traffic) "\n"
"\n"
"Disregards, " DRM(Transit.disregard) ", all Transits in this\n"
"Traffic instance."
)},

	{"force", (PyCFunction) traffic_force, METH_NOARGS,
		PyDoc_STR(
":returns: The Traffic instance being forced. (Method Chaining)\n"
":rtype: " DRC(Traffic) "\n"
"\n"
"Causes the next traffic cycle not *wait* for events. If a cycle has been started\n"
"and is currently waiting for events, cause it to stop waiting for events.\n"
)},

	{"transfer",
		(PyCFunction) traffic_transfer, METH_NOARGS,
		PyDoc_STR(
":returns: An iterator producing the In Traffic Transits, "
DRC(Transit) ", that have events.\n"
":rtype: :py:class:`collections.Iterable`\n"
"\n"
)},

	{"__enter__", (PyCFunction) traffic_enter, METH_NOARGS,
		PyDoc_STR(
"Begin a Traffic cycle."
)},

	{"__exit__", (PyCFunction) traffic_exit, METH_VARARGS,
		PyDoc_STR(
"Exit the Traffic cycle."
)},

	{NULL,},
};

static PyMemberDef traffic_members[] = {
	{"volume", T_PYSSIZET, offsetof(struct Traffic, transit_count), READONLY,
		PyDoc_STR("The number of transits being managed by the Traffic instance.")},
	{NULL,},
};

static PyObj
get_traffic_transit_count(PyObj self, void *_)
{
	Traffic T = (Traffic) self;
	return(PyLong_FromLong(T->transit_count));
}

static PyObj
get_traffic_transfer_count(PyObj self, void *_)
{
	Traffic T = (Traffic) self;
	return(PyLong_FromLong(T->transfer_count));
}

static PyGetSetDef traffic_getset[] = {
	{"transit_count", get_traffic_transit_count, NULL,
		PyDoc_STR("The number of Transits acquired by the Traffic instance.")},
	{"transfer_count", get_traffic_transfer_count, NULL,
		PyDoc_STR("The number of Transits that produced events this cycle. `-1` if not in a cycle.")},
	{NULL,},
};

static void
traffic_dealloc(PyObj self)
{
	point_dealloc(self);
}

static PyObj
traffic_new(PyTypeObject *subtype, PyObj args, PyObj kw)
{
	static char *kwlist[] = {NULL,};
	PyObj rob;
	Traffic T;

	if (!PyArg_ParseTupleAndKeywords(args, kw, "", kwlist))
		return(NULL);

	rob = PyAllocate(subtype);
	if (rob == NULL)
		return(NULL);

	T = (Traffic) rob;
	INIT_TRANSIT(T);
	Point_SetKPoint(T, -1);
	Transit_IQualify(T, teq_transfer);
	Transit_XQualify(T, teq_transfer);

	T->transit_count = 1; /* traffic itself */
	T->transfer_count = 0;
	T->ksize = 0;

	T->lltransfer_ptr = (Transit) T;

	T->traffic = T;
	T->next = (Transit) T;
	T->prev = (Transit) T;

	traffic_init_kqueue(T);

	return(rob);
}

const char traffic_doc[] =
"Traffic is a Point based Transit that manages the connection to the kernel.\n"
"Transits created should be attached to a Traffic object in order to perform\n"
"transfers with those Transits.\n"
"Traffic represents the kqueue() instance."
;
TransitPyTypeObject TrafficType = {{
	PyVarObject_HEAD_INIT(NULL, 0)
	QPATH("Traffic"),				/* tp_name */
	sizeof(struct Traffic),		/* tp_basicsize */
	0,									/* tp_itemsize */
	traffic_dealloc,				/* tp_dealloc */
	NULL,								/* tp_print */
	NULL,								/* tp_getattr */
	NULL,								/* tp_setattr */
	NULL,								/* tp_compare */
	NULL,								/* tp_repr */
	NULL,								/* tp_as_number */
	NULL,								/* tp_as_sequence */
	NULL,								/* tp_as_mapping */
	NULL,								/* tp_hash */
	NULL,								/* tp_call */
	NULL,								/* tp_str */
	NULL,								/* tp_getattro */
	NULL,								/* tp_setattro */
	NULL,								/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT,	  		/* tp_flags */
	traffic_doc,					/* tp_doc */
	NULL,								/* tp_traverse */
	NULL,								/* tp_clear */
	NULL,								/* tp_richcompare */
	0,									/* tp_weaklistoffset */
	NULL,								/* tp_iter */
	NULL,								/* tp_iternext */
	traffic_methods,				/* tp_methods */
	traffic_members,				/* tp_members */
	traffic_getset,				/* tp_getset */
	(PyTypeObject*)&PointType.typ,/* tp_base */
	NULL,								/* tp_dict */
	NULL,								/* tp_descr_get */
	NULL,								/* tp_descr_set */
	0,									/* tp_dictoffset */
	NULL,								/* tp_init */
	NULL,								/* tp_alloc */
	traffic_new,					/* tp_new */
},
	&TrafficTIF,
};

const char transit_error_doc[] =
"Abstract transit error.\n"
"\n"
"All :py:mod:`traffic.libkqueue` exceptions inherit from this one.\n"
;

const char terminated_error_doc[] =
"Error raised when an operation was performed on a terminated Transit.\n"
;

const char resource_error_doc[] =
"Error raised when "DRM(Transit.acquire)" is called prior to an exhaust event.\n"
;

/* METH_O, METH_VARARGS, METH_VARKEYWORDS, METH_NOARGS */
static PyMethodDef methods[] = {
	/*{"name", (PyCFunction) func, METH_NOARGS, "doc-string"},*/
	{NULL}
};

#define TRANSIT_TYPES() \
	ID(Transit) \
	ID(Point) \
	ID(Traffic) \
	ID(Octets) \
	ID(Sockets)

#define ID(N) eid_##N = NULL, eid_##N##_str = NULL,
static PyObj TRANSIT_EVENTS()
	transit_event_names = NULL;
#undef ID

#define EVENT_COMBINATIONS \
	EC(1<<tev_terminate, t) \
	EC(1<<tev_transfer, f) \
	EC(1<<tev_exhaust, x) \
	EC(1<<tev_terminate|1<<tev_transfer, ft) \
	EC(1<<tev_transfer|1<<tev_exhaust, fx)

#define EC(bits, name) EventSet_##name = NULL,
static PyObj
EVENT_COMBINATIONS
EventSet_None = NULL;
#undef EC

static PyObj
select_events(Transit t)
{
	PyObj rob;

	switch (t->event)
	{
		case 0:
			rob = EventSet_None;
		break;

#define EC(casev, name) \
		case casev: \
			rob = EventSet_##name; \
		break;

EVENT_COMBINATIONS

#undef EC
	}

	Py_INCREF(rob);
	return(rob);
}

/*
 * initialize the EventSet_* frozensets.
 */
static int
init_event_sets(void)
{
	PyObj x;

	x = PyTuple_New(2); if (x == NULL) return (-1);

	PyTuple_SET_ITEM(x, 0, TRAFFIC_EVENT_OB(tev_transfer));
	Py_INCREF(TRAFFIC_EVENT_OB(tev_transfer));
	PyTuple_SET_ITEM(x, 1, TRAFFIC_EVENT_OB(tev_exhaust));
	Py_INCREF(TRAFFIC_EVENT_OB(tev_exhaust));

	EventSet_fx = PyFrozenSet_New(x);
	Py_DECREF(x);
	if (EventSet_fx == NULL) return (-1);

	x = PyTuple_New(2); if (x == NULL) return (-1);

	PyTuple_SET_ITEM(x, 0, TRAFFIC_EVENT_OB(tev_terminate));
	Py_INCREF(TRAFFIC_EVENT_OB(tev_terminate));
	PyTuple_SET_ITEM(x, 1, TRAFFIC_EVENT_OB(tev_transfer));
	Py_INCREF(TRAFFIC_EVENT_OB(tev_transfer));

	EventSet_ft = PyFrozenSet_New(x);
	Py_DECREF(x);
	if (EventSet_ft == NULL) return (-1);

	x = PyTuple_New(1); if (x == NULL) return (-1);

	PyTuple_SET_ITEM(x, 0, TRAFFIC_EVENT_OB(tev_terminate));
	Py_INCREF(TRAFFIC_EVENT_OB(tev_terminate));

	EventSet_t = PyFrozenSet_New(x);
	Py_DECREF(x);
	if (EventSet_t == NULL) return (-1);


	x = PyTuple_New(1); if (x == NULL) return (-1);

	PyTuple_SET_ITEM(x, 0, TRAFFIC_EVENT_OB(tev_exhaust));
	Py_INCREF(TRAFFIC_EVENT_OB(tev_exhaust));

	EventSet_x = PyFrozenSet_New(x);
	Py_DECREF(x);
	if (EventSet_x == NULL) return (-1);


	x = PyTuple_New(1); if (x == NULL) return (-1);

	PyTuple_SET_ITEM(x, 0, TRAFFIC_EVENT_OB(tev_transfer));
	Py_INCREF(TRAFFIC_EVENT_OB(tev_transfer));

	EventSet_f = PyFrozenSet_New(x);
	Py_DECREF(x);
	if (EventSet_f == NULL) return (-1);

	EventSet_None = PyFrozenSet_New(NULL);
	if (EventSet_None == NULL) return(-1);

	return(0);
}

/*
 * Build out the array.array("i", (-1,)).__mul__ object
 * for use in Sockets.rallocate(n)
 */
static PyObj
mk_new_array(void)
{
	PyObj rob = NULL; /* array.array("i", (-1,)).__mul__ */
	PyObj args, typ, vals; /* args to array.array() */
	PyObj mod, at, ai; /* array mod, array type, array instance */

	args = PyTuple_New(2);
	if (args == NULL)
		return(NULL);

	typ = PyUnicode_FromString("i");
	if (typ == NULL)
		goto d_args;
	PyTuple_SET_ITEM(args, 0, typ);

	vals = PyTuple_New(1);
	if (vals == NULL)
		goto d_args;
	PyTuple_SET_ITEM(args, 1, vals);

	PyTuple_SET_ITEM(vals, 0, PyLong_FromLong(-1));
	if (PyTuple_GET_ITEM(vals, 0) == NULL)
		goto d_args;

	mod = PyImport_ImportModule("array");
	if (mod == NULL)
		goto d_args;

	at = PyObject_GetAttrString(mod, "array");
	if (at == NULL)
		goto d_mod;

	ai = PyObject_CallObject(at, args);
	if (ai == NULL)
		goto d_at;

	rob = PyObject_GetAttrString(ai, "__mul__");

d_ai:
	Py_DECREF(ai);
d_at:
	Py_DECREF(at);
d_mod:
	Py_DECREF(mod);
d_args:
	Py_DECREF(args);
	return(rob);
}

#if PY_MAJOR_VERSION > 2
static struct PyModuleDef module = {
   PyModuleDef_HEAD_INIT,
   MODULE,		/* name of module */
   NULL, 		/* module documentation, may be NULL */
   -1,			/* size of per-interpreter state of the module,
						or -1 if the module keeps state in global
						variables. */
   methods,		/* methods */
};
#endif

PyMODINIT_FUNC
MODULE_ENTRY_POINT(libkqueue)(void)
{
	PyObj mod = NULL;

	if (new_array == NULL)
	{
		/*
		 * used by Sockets.rallocate(n) to build
		 */
		new_array = mk_new_array();
		if (new_array == NULL)
#if PY_MAJOR_VERSION > 2
			return(NULL);
#else
			return;
#endif
	}

#if PY_MAJOR_VERSION > 2
	mod = PyModule_Create(&module);
	if (mod == NULL)
		return(NULL);
#else
	mod = Py_InitModule(MODULE, methods);
	if (mod == NULL)
		return;
#endif

	if (PyExc_TransitError == NULL)
	{
		PyExc_TransitError = PyErr_NewExceptionWithDoc(
			QPATH("TransitError"), transit_error_doc, PyExc_Exception, NULL);
		if (PyExc_TransitError == NULL)
			goto cleanup;
	}

	if (PyExc_TerminatedError == NULL)
	{
		PyExc_TerminatedError = PyErr_NewExceptionWithDoc(
			QPATH("TerminatedError"), terminated_error_doc, PyExc_TransitError, NULL);
		if (PyExc_TerminatedError == NULL)
			goto cleanup;
	}

	if (PyExc_ResourceError == NULL)
	{
		PyExc_ResourceError = PyErr_NewExceptionWithDoc(
			QPATH("ResourceError"), resource_error_doc, PyExc_TransitError, NULL);
		if (PyExc_ResourceError == NULL)
			goto cleanup;
	}

	Py_XDECREF(transit_event_names);
	transit_event_names = NULL;
	transit_event_names = PyDict_New();
	if (transit_event_names == NULL)
		goto cleanup;

	if (PyType_Ready(&(lltransferiter_type)))
		goto cleanup;
	if (PyType_Ready(&(eventiter_type)))
		goto cleanup;

	/*
	 * Initialize exceptions on the Transit dictionary
	 */
#define ID(NAME) \
	if (PyModule_AddObject(mod, #NAME, PyExc_##NAME) < 0) \
		goto cleanup;
	TRANSIT_EXCEPTIONS()
#undef ID

	/*
	 * Build the KC string objects for the failure() method.
	 */
#define KC(x) \
	if (kc_##x##_ob == NULL) \
		{ kc_##x##_ob = PyUnicode_FromString(#x); \
			if (kc_##x##_ob == NULL) goto cleanup; }
	KCALLS()
#undef KC

#define ID(NAME) \
	if (eid_##NAME##_str == NULL) \
		{ eid_##NAME##_str = PyUnicode_FromString(#NAME); \
			if (eid_##NAME##_str == NULL) goto cleanup; } \
	if (eid_##NAME == NULL) { \
		eid_##NAME = PyLong_FromLong(tev_##NAME); \
		if (eid_##NAME == NULL) goto cleanup; }
	TRANSIT_EVENTS()
#undef ID

#define ID(NAME) TRAFFIC_EVENT_OB(tev_##NAME) = eid_##NAME;
	TRANSIT_EVENTS()
#undef ID

#define ID(NAME) PyDict_SetItem(transit_event_names, eid_##NAME, eid_##NAME##_str);
	TRANSIT_EVENTS()
#undef ID
	PyModule_AddObject(mod, "event_names", transit_event_names);

	if (init_event_sets() != 0)
		goto cleanup;

	/*
	 * Initialize Transit types.
	 */
#define ID(NAME) \
	if (PyType_Ready((PyTypeObject *) &( NAME##Type ))) \
		goto cleanup; \
	PyModule_AddObject(mod, #NAME, (PyObj) &( NAME##Type ));
	TRANSIT_TYPES()
#undef ID

#if PY_MAJOR_VERSION > 2
	return(mod);
cleanup:
	Py_DECREF(mod);
	return(NULL);
#else
cleanup:
	return;
#endif
}
/*
 * vim: ts=3:sw=3:noet:
 */
