"""
Tools for working with :py:class:`traffic.libkqueue.Traffic` cycles.
"""
import operator
import collections

#: Singleton Tuple of event identifiers produced by :py:class:`traffic.libkqueue.Transit`
#: instances. Useful for referencing event identifiers without importing the
#: primary library.
events = collections.namedtuple("Events", "terminate transfer exhaust")(*range(3))

def unroll(traffic):
	"""
	unroll(traffic)

	:param traffic: The traffic instance whose window should be unrolled.
	:rtype: generator
	:returns: An iterator producing transit-event pairs: (transit, event).

	Given a :py:class:`traffic.libkqueue.Traffic` instance, yield each event that
	occurred associated with the Transit that produced the event.
	"""
	for t in traffic.transfer():
		for e in t.events():
			yield (t, e)

#: Used to hold a snapshot of events and data that were produced by a Transit and their
#: associated information or callback. This is useful for extracting transfer
#: data from a Transit that can be used outside of the Traffic Cycle.
#: 
#: Attributes:
#: 
#:  **transit**
#:   The :py:class:`traffic.libkqueue.Transit` instance.
#: 
#:  **link**
#:   The :py:attr:`traffic.libkqueue.Transit.link` attribute of the `transit`.
#:   In cases of termination, the Transit will drop the `link` attribute after
#:   the terminating cycle.
#: 
#:  **terminated**
#:   Whether or not a terminate event occurred: `None` if not terminated.
#:   Otherwise, the result of the :py:meth:`traffic.libkqueue.Point.failure` method.
#: 
#:  **transferred**
#:   An integer designating the amount transferred iff the transit sends.
#:   The transfer slice of the resource iff the transit receives.
#: 
#:  **demand**
#:   The :py:meth:`traffic.libkqueue.Transit.acquire` method if an exhaust event
#:   occurred, otherwise, `None`.
#:
Activity = collections.namedtuple("Activity",
	("transit", "link", "termination", "transferred", "demand",)
)

def activity(transit,
	term_event = events.terminate,
	xfer_event = events.transfer,
	exht_event = events.exhaust,
	Activity = Activity,
):
	"""
	Given a :py:class:`traffic.libkqueue.Transit`, get an :py:class:`traffic.libloop.Activity`
	snapshot of the Transit.
	"""
	termination = None # not terminated
	demand = None
	xfer = None

	for event in transit.events():
		if event == xfer_event:
			xfer = transit.resource[transit.transfer()]
		elif event == exht_event:
			demand = transit.acquire
			# termination and exhaust follow transfers
			# termination cannot follow exhaust
			break
		else:
			termination = transit.failure()
			# termination and exhaust follow transfers
			# exhaust cannot follow termination
			break

	return Activity(transit, transit.link, termination, xfer, demand)

def snapshot(traffic, _list = list, _map = map):
	"""
	:param traffic: The Traffic instance to extract Activity from.
	:type traffic: :py:class:`traffic.libkqueue.Traffic`
	:returns: List of Activity instances.
	:rtype: [:py:class:`Activity`]

	Given a :py:class:`traffic.libkqueue.Traffic` instance, get a snapshot of all
	:py:func:`traffic.libloop.activity`.
	"""
	# list() is invoked here as materialization of the
	# snapshot is needed in cases where the cycle exits. (like libloop.cycle)
	return _list(_map(activity, traffic.transfer()))

def cycle(traffic):
	"""
	:param traffic: The Traffic to cycle.
	:type traffic: :py:class:`traffic.libkqueue.Traffic`
	:returns: List of Activity instances.
	:rtype: [:py:class:`Activity`]

	Cycle the traffic instance and return a :py:func:`traffic.libloop.snapshot`
	list of activity.

	.. warning:: If insufficient memory is available for building the snapshot,
	             transfer-loss *will* occur.
	"""
	with traffic:
		return snapshot(traffic)

def loop(deliver, traffic, _cycle = cycle):
	"""
	:param deliver: The callable that is given the sequence of :py:class:`Activity` instances.
	:type deliver: callable
	:param traffic: The Traffic instance that is being cycled until termination.
	:type traffic: :py:class:`traffic.libkqueue.Traffic`
	:returns: None

	Function that runs a Traffic loop and delivers sequences of
	:py:class:`Activity` instances into the given :py:obj:`deliver` callable
	so long as the Traffic instance is not terminated::

		Q = queue.Queue(8)
		T = libtraffic.Traffic()
		libloop.loop(Q.put, T)

	Running the call to loop in a thread is an appropriate way to collect events.
	By doing this, events and wire transfers can be processed in parallel.
	"""
	while not traffic.terminated:
		deliver(_cycle(traffic))
