##
# .test.test_libtraffic
##
import time
import os
import socket
import struct
from .. import libkqueue as libtraffic
from .. import libloop

def test_module_protocol(test):
	test.fail_if_not_hasattr(libtraffic, "Transit")
	test.fail_if_not_hasattr(libtraffic, "Point")
	test.fail_if_not_hasattr(libtraffic, "Traffic")
	test.fail_if_not_hasattr(libtraffic, "Octets")
	test.fail_if_not_hasattr(libtraffic, "Sockets")

	test.fail_if_not_subclass(libtraffic.Point, libtraffic.Transit)
	test.fail_if_not_subclass(libtraffic.Traffic, libtraffic.Point)
	test.fail_if_not_subclass(libtraffic.Octets, libtraffic.Point)
	test.fail_if_not_subclass(libtraffic.Sockets, libtraffic.Point)
	types = (
		libtraffic.Transit,
		libtraffic.Point, libtraffic.Traffic,
		libtraffic.Octets, libtraffic.Sockets,
	)

def test_subtyping(test):
	types = (
		libtraffic.Traffic, libtraffic.Octets, libtraffic.Sockets,
	)
	for x in types:
		with test.trap(TypeError) as t:
			# Transit types extend the storage internally.
			# Discourage subtyping.
			# XXX: Transit can still be subclassed?
			class Foo(x):
				pass

def test_sockets_rallocate(test):
	test.fail_if_not_equal(list(libtraffic.Sockets.rallocate(10)), [-1] * 10)
	sb = libtraffic.Sockets.rallocate(16)
	mv = memoryview(sb)
	test.fail_if_equal(sb[0], 1) # should be -1, but anything aside from 1 is okay.
	# itemsize is 4, so use pack_into.
	struct.pack_into("i", mv, 0, 1)
	test.fail_if_not_equal(sb[0], 1)

def test_octets_rallocate(test):
	test.fail_if_not_equal(list(libtraffic.Octets.rallocate(20)), [b'\x00'] * 20)
	mb = libtraffic.Octets.rallocate(20)
	mb[0:5] = b'fffff'
	test.fail_if_not_equal(memoryview(mb).tobytes(), b'fffff' + (b'\x00' * (20 - 5)))
	mv = memoryview(mb)
	b = bytes(mv[11])
	mb[10:15] = b'fffff'

def test_polarity(test):
	(r1,w1),(r2,w2) = libtraffic.Octets.connect_stream_socketpair()
	test.fail_if_in(False, set((r1.polarity, r2.polarity)))
	test.fail_if_in(True, set((w1.polarity, w2.polarity)))

def test_traffic_termination(test):
	'traffic termination sequence with no transits'
	T = libtraffic.Traffic()
	test.fail_if(T.terminated)
	T.terminate()
	test.fail_if(T.terminated) # it hasn't completed termination.
	with T:
		pass
	# one cycle in..
	test.fail_if_not(T.terminated)
	test.fail_if_not_raised(libtraffic.TerminatedError, T.__enter__)

def test_traffic_exit(test):
	'context manager to terminate on exit'
	class LocalExc(Exception):
		pass
	T = libtraffic.Traffic()
	T.force()
	with test.trap(LocalExc) as trap:
		with T:
			raise LocalExc("foo")
	test.fail_if(T.terminated)

def test_traffic_force(test):
	T = libtraffic.Traffic()
	# once per second
	# this gives T a kevent
	T.force()
	start = time.time()
	with T:
		period = time.time() - start
		test.fail_if_not_empty(T.transfer())
	# rather aribtrary number...
	test.fail_if_greater_than(period, 0.3)
	T.terminate()

def test_multitraffic(test, number_to_check = 128):
	'Validate that multiple Traffic instances can exist.'
	traffics = [libtraffic.Traffic() for x in range(number_to_check)]
	for x in traffics:
		test.fail_if_not_equal(x._errno, 0)
		x.terminate()
		test.fail_if_not_equal(x._errno, 0)
		with x:
			test.fail_if_not_equal(x._errno, 0)
			pass
		test.fail_if_not_equal(x._errno, 0)

def traffic_termination(test, T):
	T.terminate()
	# traffic termination cascades to transits
	with T:
		transits = set(T.transfer())
	test.fail_if_not(T.terminated)
	for x in transits:
		test.fail_if_not(x.terminated)
	return transits

def connection_transfer(test, T, client, server):
	data = [b'foobar', b'barfoo']

	client[1].acquire(data[0])
	server[1].acquire(data[1])

	# do writes
	# NOTE: assumes receive buffers are large enough for 'foobar' ;)
	exhausted = 0
	while exhausted != 2:
		for x in libloop.cycle(T):
			if x.demand is not None:
				exhausted += 1

	# do reads
	expected = list(reversed(data))
	buffers = [bytearray(), bytearray()]
	idxmap = {client[0]: 0, server[0]: 1}

	client[0].acquire(client[0].rallocate(6))
	server[0].acquire(server[0].rallocate(6))

	exhausted = 0
	while exhausted != 2:
		for x in libloop.cycle(T):
			if x.transferred:
				buffers[idxmap[x.transit]] += x.transferred
			if x.demand:
				exhausted += 1

	test.fail_if_not_equal(expected, buffers)

def connection_termination(test, T, client, server):
	'checked that remote termination is signalling termination on a receiver'

	client_side = (client[0], server[1])
	server_side = (server[0], client[1])
	for term in (client_side, server_side):
		term[1].terminate()

		# collect term events
		acts = list()
		while len(acts) < 2:
			for x in libloop.cycle(T):
				if x.termination is not None:
					acts.append(x)

		# validate that they are the expected transits
		# and terminateD status
		for x in acts:
			test.fail_if_not_in(x.transit, term)
			test.fail_if_not(x.transit.terminated)

def new_socketpair_transfer(test, T):
	spairs = libtraffic.Octets.connect_stream_socketpair()
	for x in spairs[0] + spairs[1]:
		T.acquire(x)
	return connection_transfer(test, T, spairs[0], spairs[1])

def test_socketpair_transfer(test):
	T = libtraffic.Traffic()
	new_socketpair_transfer(test, T)
	transits = traffic_termination(test, T)

def new_pipe_transfer(test, T):
	r1, w2 = libtraffic.Octets.connect_pipe()
	r2, w1 = libtraffic.Octets.connect_pipe()
	for x in (r1,w1,r2,w2):
		T.acquire(x)
	connection_transfer(test, T, (r1, w1), (r2, w2))
	return {r1,r2,w1,w2}

def test_octets_resource_error(test):
	r, w = libtraffic.Octets.connect_pipe()
	# needs mutable buffer
	test.fail_if_not_raised(BufferError, r.acquire, b'')
	# already acquired, writer
	w.acquire(b'')
	test.fail_if_not_raised(libtraffic.ResourceError, w.acquire, b'')
	# already acquired, reader
	r.acquire(r.rallocate(0))
	test.fail_if_not_raised(libtraffic.ResourceError, r.acquire, r.rallocate(0))

def test_pipe(test):
	T = libtraffic.Traffic()
	created = new_pipe_transfer(test, T)
	traffic_termination(test, T)
	for x in created:
		test.fail_if_not_equal(True, x.terminated)

def test_connect_io_point(test):
	'test arbitrary '
	T = libtraffic.Traffic()
	xr1, xw2 = os.pipe()
	xr2, xw1 = os.pipe()
	r1 = libtraffic.Octets.connect_input_point(xr1)
	w1 = libtraffic.Octets.connect_output_point(xw1)
	r2 = libtraffic.Octets.connect_input_point(xr2)
	w2 = libtraffic.Octets.connect_output_point(xw2)
	T.acquire(r1)
	T.acquire(r2)
	T.acquire(w1)
	T.acquire(w2)
	connection_transfer(test, T, (r1, w1), (r2, w2))
	traffic_termination(test, T)
	test.fail_if_not_equal(True, r1.terminated)
	test.fail_if_not_equal(True, r2.terminated)
	test.fail_if_not_equal(True, w1.terminated)
	test.fail_if_not_equal(True, w2.terminated)

def test_traffic_flush_irelease(test):
	"Validates the Transit's resource is released on flush"
	T = libtraffic.Traffic()
	r, w = libtraffic.Octets.connect_pipe()
	T.acquire(r)
	T.acquire(w)
	w.acquire(b'1')
	r.acquire(bytearray(1))
	exhausted = 0
	while exhausted < 2:
		for x in libloop.cycle(T):
			if x.demand is not None:
				exhausted += 1
	# New resource is not acquired in loop, but exhaust
	# needs to irelease the resource in the __exit__/flush.
	test.fail_if_not_exact(r.resource, None)
	test.fail_if_not_exact(w.resource, None)

	# now the termination branch
	w.acquire(b'1')
	r.acquire(bytearray(1))
	test.fail_if_exact(r.resource, None)
	test.fail_if_exact(w.resource, None)
	r.terminate()
	w.terminate()
	terms = 0
	while terms < 2:
		for x in libloop.cycle(T):
			if x.termination is not None:
				terms += 1
			# also fail if there is a demand reference.
			test.fail_if_not_exact(x.demand, None)
	# validate that termination causes the resources to be released.
	test.fail_if_not_exact(r.resource, None)
	test.fail_if_not_exact(w.resource, None)

# Marked as broken in the documentation. EVFILT_READ/WRITE with files
# behaves strangely at best and required some special casing in Octets.
# This test passes on Lion, but I suspect kqueue()'s usage with regular
# files should not be relied on.
def test_file(test):
	import os, tempfile, shutil
	tmpdir = tempfile.mkdtemp()
	try:
		path = os.path.join(tmpdir, "wfile")
		apath = path.encode('ascii')
		count = 0
		wrote_data = []
		thedata = b'\xF1'*128

		T = libtraffic.Traffic()
		wr = libtraffic.Octets.connect_file_append(apath)
		T.acquire(wr)
		wr.acquire(thedata)
		test.fail_if_not_exact(wr.resource, thedata)
		while not wr.terminated:
			with T:
				for x in T.transfer():
					for e in x.events():
						if e == libloop.events.transfer:
							pass
						elif e == libloop.events.exhaust:
							wrote_data.append(thedata)
							x.acquire(thedata)
							count += 1
							if count == 16:
								x.terminate()
		os.chmod(apath, 0o700)
		# now read it back in.
		data = []
		rd = libtraffic.Octets.connect_file(apath)
		T.acquire(rd)
		rd.acquire(bytearray(128))
		terminating = False
		while not rd.terminated:
			with T:
				for x in T.transfer():
					for e in x.events():
						if e == libloop.events.transfer:
							data.append(bytes(x.resource[x.transfer()]))
							pass
							if sum(map(len,data)) == len(thedata):
								x.terminate()
								terminating = True
						elif e == libloop.events.exhaust and not terminating:
							x.acquire(bytearray(128))
	finally:
		shutil.rmtree(tmpdir)

def showevents(T):
	transits = tuple(T.transfer())
	if not transits:
		print("no events")
	else:
		for x in transits:
			print(x, x.identity(), x.opposite.identity() if hasattr(x, 'opposite') else None, ":", tuple(x.events()))

def accept_stream(test, T, sf, client):
	sf.acquire(sf.rallocate(2))
	for x in client:
		T.acquire(x)

	server = None
	while server is None:
		# loop to help ensure that the connection happens
		ca = libloop.cycle(T)
		for x in ca:
			x = x.transferred[0]
			server = tuple(map(T.acquire, libtraffic.Octets.connect_stream_socket(x)))

	libloop.cycle(T)
	for x in client + server:
		T.acquire(x)
	connection_transfer(test, T, client, server)
	# exhausted, nothing should be happening
	connection_termination(test, T, client, server)

def test_sockets_stream_local(test):
	import os, tempfile, shutil
	tmpdir = tempfile.mkdtemp()
	afilepath = os.path.join(tmpdir, 'server_fss')
	filepath = afilepath.encode('ascii')
	try:
		sf = libtraffic.Sockets.accept_stream_local(filepath)
		os.chmod(filepath, 0o700)
		T = libtraffic.Traffic()
		T.acquire(sf)
		test.fail_if_not_equal(filepath, sf.identity()[0])
		client = libtraffic.Octets.connect_stream_local(filepath)
		test.fail_if_not_equal(filepath, client[1].identity()[0])

		accept_stream(test, T, sf, client)
	finally:
		shutil.rmtree(tmpdir)

def test_sockets_stream_ip4(test):
	sf = libtraffic.Sockets.accept_stream_ip4(interface = ('127.0.0.1', 0))
	T = libtraffic.Traffic()
	T.acquire(sf)
	host, port = sf.identity() # zero port, OS selects available port
	test.fail_if_not_equal(host, '127.0.0.1')

	client = libtraffic.Octets.connect_stream_ip4(('127.0.0.1', port))
	accept_stream(test, T, sf, client)

def test_failure_on_bind(test):
	sf = libtraffic.Sockets.accept_stream_ip4(interface = ('127.0.0.1', 0))
	T = libtraffic.Traffic()
	T.acquire(sf)
	host, port = sf.identity() # zero port, OS selects available port
	fail = libtraffic.Sockets.accept_stream_ip4(interface = (host, port))
	test.fail_if_not_equal(fail.failure(), ("bind", 48))
	test.fail_if_not_equal(fail.terminated, False) # hasn't been terminated yet.
	T.acquire(fail)
	r = libloop.cycle(T)
	test.fail_if_not(r) # expecting a terminate from fail
	test.fail_if_not_equal(fail.failure(), ("bind", 48))
	test.fail_if_not_equal(fail.terminated, True)
	fa = r[0]
	test.fail_if_not_equal(fa.termination, ("bind", 48))
	T.terminate()
	while not sf.terminated:
		libloop.cycle(T)
	test.fail_if_not_equal(T.terminated, True)

def test_sockets_stream_ip6(test):
	sf = libtraffic.Sockets.accept_stream_ip6(('::1', 0))
	T = libtraffic.Traffic()
	T.acquire(sf)
	host, port = sf.identity()
	test.fail_if_not_equal(host, '::1')

	client = tuple(map(T.acquire, libtraffic.Octets.connect_stream_ip6((host,port))))
	accept_stream(test, T, sf, client)

def test_sockets_stream_socket(test):
	# essentially repeating the ip4 test
	s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	s.bind(('127.0.0.1', 0))
	fd = os.dup(s.fileno())
	s.close()

	sf = libtraffic.Sockets.accept_stream_socket(fd)
	T = libtraffic.Traffic()
	T.acquire(sf)
	host, port = sf.identity() # zero port, OS selects available port
	test.fail_if_not_equal(host, '127.0.0.1')

	client = libtraffic.Octets.connect_stream_ip4(('127.0.0.1', port))
	accept_stream(test, T, sf, client)

def test_octets_datagram_connect_ip4(test):
	# same code path as stream, so just check instantiation
	T = libtraffic.Traffic()
	r, w = libtraffic.Octets.connect_datagram_ip4(('127.0.0.1', 1))
	T.acquire(r)
	T.acquire(w)
	r.terminate()
	w.terminate()
	libloop.cycle(T)
	T.terminate()

def test_octets_datagram_connect_ip6(test):
	# same code path as stream, so just check instantiation
	T = libtraffic.Traffic()
	r, w = libtraffic.Octets.connect_datagram_ip4(('::1', 1))
	T.acquire(r)
	T.acquire(w)
	r.terminate()
	w.terminate()
	libloop.cycle(T)
	T.terminate()

def test_unexpected_close(test):
	T = libtraffic.Traffic()
	client, server = libtraffic.Octets.connect_stream_socketpair()
	for x in client + server:
		T.acquire(x)
	os.close(T._fileno)
	test.fail_if_not_equal(libloop.cycle(T), [])
	test.fail_if(T.terminated)

	connection_transfer(test, T, client, server)
	traffic_termination(test, T)

def test_unexpected_close_in_cycle(test):
	T = libtraffic.Traffic()
	client, server = libtraffic.Octets.connect_stream_socketpair()
	for x in client + server:
		T.acquire(x)
	T.__enter__()
	os.close(T._fileno)
	T.__exit__(None, None, None)
	test.fail_if_not_equal(libloop.cycle(T), [])
	test.fail_if(T.terminated)

	connection_transfer(test, T, client, server)
	traffic_termination(test, T)

def test_disregard_detached(test):
	"checks disregard's for Transits *not* in Traffic"
	# shares the FD
	r, w = libtraffic.Octets.connect_stream_socketpair()
	# doesn't share the FD
	p = libtraffic.Octets.connect_pipe()

	fds = set()
	for x in r + w + p:
		fds.add(x._fileno)
		test.fail_if_not_exact(x.disregard(), x)
		test.fail_if_not_equal(x._fileno, -1)

	# have to check afterwards because of Octets_IsSocket()
	# has a two-phase close to handled shared the file descriptor.
	for x in fds:
		test.fail_if_not_raised(OSError, os.close, x)

def test_disregard_attached(test):
	"checks disregard's for Transits in Traffic"
	# shares the FD
	r, w = libtraffic.Octets.connect_stream_socketpair()
	# doesn't share the FD
	p = libtraffic.Octets.connect_pipe()
	T = libtraffic.Traffic()

	fds = set()
	for x in r + w + p:
		T.acquire(x)
		fds.add(x._fileno)

	test.fail_if_not_exact(x.disregard(), x)
	test.fail_if_not_equal(x._fileno, -1)

	test.fail_if_not_exact(T.disregard(), T)

	# all of the FDs should be close(2)'d
	for x in fds:
		test.fail_if_not_raised(OSError, os.close, x)

	# now produce the terminate events
	events = libloop.cycle(T)
	test.fail_if_not_equal(len(events), len(r + w + p))
	for x in events:
		test.fail_if_not_equal(x.termination, (None, 0))

if __name__ == '__main__':
	from dev import libtest; libtest.execmodule()
